"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from qrisp import QuantumFloat
import numpy as np


def prepare(qv, target_array, reversed=False):
    r"""
    This method performs quantum state preparation. Given a vector $b=(b_0,\dotsc,b_{N-1})$, the function acts as

    .. math::

        \ket{0} \rightarrow \sum_{i=0}^{N-1}b_i\ket{i}

    Parameters
    ----------
    qv : QuantumVariable
        The quantum variable on which to apply state preparation.
    target_array : numpy.ndarray
        The vector $b$.
    reversed : boolean
        If set to ``True``, the endianness is reversed. The default is ``False``.

    Examples
    --------

    We create a :ref:`QuantumFloat` and prepare the state $\sum_{i=0}^3b_i\ket{i}$ for $b=(0,1,2,3)$.

    ::

        b = np.array([0,1,2,3])

        qf = QuantumFloat(2)
        prepare(qf, b)

        res_dict = qf.get_measurement()

        for k, v in res_dict.items():
            res_dict[k] = v**0.5

        for k, v in res_dict.items():
            res_dict[k] = v/res_dict[1.0]

        print(res_dict)
        # Yields: {3: 2.9999766670425863, 2: 1.999965000393743, 1: 1.0}

    """

    from qiskit.circuit.library.data_preparation.state_preparation import (
        StatePreparation,
    )

    n = len(target_array)
    m = int(np.ceil(np.log2(n)))

    # Fill target_array with zeros
    if not (n & (1 << m)):
        target_array = np.concatenate((target_array, np.zeros((1 << m) - n)))

    target_array = target_array / np.vdot(target_array, target_array) ** 0.5

    qiskit_qc = StatePreparation(target_array).definition
    from qrisp import QuantumCircuit

    init_qc = QuantumCircuit.from_qiskit(qiskit_qc)

    # Find global phase correction
    from qrisp.simulator import statevector_sim

    init_qc.qubits.reverse()
    sim_array = statevector_sim(init_qc)
    init_qc.qubits.reverse()

    arg_max = np.argmax(np.abs(sim_array))

    gphase_dif = (np.angle(target_array[arg_max] / sim_array[arg_max])) % (2 * np.pi)

    init_qc.gphase(gphase_dif, 0)

    init_gate = init_qc.to_gate()

    init_gate.name = "state_init"

    if reversed:
        qv.qs.append(init_gate, [qv[m - 1 - i] for i in range(m)])
    else:
        qv.qs.append(init_gate, [qv[i] for i in range(m)])



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from qrisp.interface.virtual_backend import VirtualBackend


class QiskitBackend(VirtualBackend):
    """
    This class instantiates a :ref:`VirtualBackend` using a Qiskit backend.
    This allows easy access to Qiskit backends through the qrisp interface.

    Parameters
    ----------
    backend : Qiskit backend object, optional
        A Qiskit backend object, which runs QuantumCircuits. The default is
        ``AerSimulator()``.
    port : int, optional
        The port to listen. The default is None.

    Examples
    --------

    We evaluate a :ref:`QuantumFloat` multiplication on the Aer simulator.

    >>> from qrisp import QuantumFloat
    >>> from qrisp.interface import QiskitBackend
    >>> from qiskit_aer import AerSimulator
    >>> example_backend = QiskitBackend(backend = AerSimulator())
    >>> qf = QuantumFloat(4)
    >>> qf[:] = 3
    >>> res = qf*qf
    >>> res.get_measurement(backend = example_backend)
    {9: 1.0}

    We evaluate a :ref:`QuantumFloat` multiplication on the FakeWashingtonV2 backend.

    >>> from qrisp import QuantumFloat
    >>> from qiskit_ibm_runtime.fake_provider import FakeWashingtonV2
    >>> from qrisp.interface import QiskitBackend
    >>> example_backend = QiskitBackend(backend = FakeWashingtonV2())
    >>> qf = QuantumFloat(2)
    >>> qf[:] = 2
    >>> res = qf*qf
    >>> res.get_measurement(backend = example_backend)
    {4: 0.6962,
    12: 0.0967,
    0: 0.0607,
    8: 0.0572,
    6: 0.028,
    2: 0.0128,
    14: 0.0126,
    5: 0.0103,
    10: 0.0062,
    3: 0.0057,
    9: 0.0042,
    13: 0.0037,
    1: 0.0029,
    7: 0.001,
    15: 0.001,
    11: 0.0008}

    We evaluate a :ref:`QuantumFloat` addition on a real IBM quantum backend.

    >>> from qrisp import QuantumFloat
    >>> from qrisp.interface import QiskitBackend
    >>> from qiskit_ibm_runtime import QiskitRuntimeService
    >>> service = QiskitRuntimeService(channel="ibm_cloud", token="YOUR_IBM_CLOUD_TOKEN")
    >>> brisbane = service.backend("ibm_brisbane")
    >>> qrisp_brisbane = QiskitBackend(backend)
    >>> qf = QuantumFloat(2)
    >>> qf[:] = 2
    >>> qf+=1
    >>> qf.get_measurement(backend = qrisp_brisbane)
    {3: 0.919, 1: 0.044, 2: 0.021, 0: 0.016}

    """

    def __init__(self, backend=None, port=None):

        if backend is None:
            try:
                from qiskit_aer import AerSimulator
                backend = AerSimulator()
            except ImportError:
                raise ImportError("Encountered ImportError when trying to import AerSimulator. Likely caused by incompatible qiskit and qiskit-aer versions.")

        try:
            from qiskit_ibm_runtime import SamplerV2
        except ImportError:
            raise ImportError("Please install qiskit-ibm-runtime to use the QiskitBackend. You can do this by running `pip install qiskit-ibm-runtime`.")
        sampler = SamplerV2(backend)

        # Create the run method
        def run(qasm_str, shots=None, token=""):
            if shots is None:
                shots = 1000
            # Convert to qiskit
            from qiskit import QuantumCircuit

            qiskit_qc = QuantumCircuit.from_qasm_str(qasm_str)

            # Make circuit with one monolithic register
            new_qiskit_qc = QuantumCircuit(len(qiskit_qc.qubits), len(qiskit_qc.clbits))
            for instr in qiskit_qc:
                new_qiskit_qc.append(
                    instr.operation,
                    [qiskit_qc.qubits.index(qb) for qb in instr.qubits],
                    [qiskit_qc.clbits.index(cb) for cb in instr.clbits],
                )

            from qiskit import transpile

            qiskit_qc = transpile(new_qiskit_qc, backend=backend)

            job = sampler.run([qiskit_qc], shots=shots)     

            qiskit_result = (
                job.result()[0].data.c.get_counts() 
                # https://docs.quantum.ibm.com/migration-guides/v2-primitives
            )

            # Remove the spaces in the qiskit result keys
            result_dic = {}
            import re

            for key in qiskit_result.keys():
                counts_string = re.sub(r"\W", "", key)
                result_dic[counts_string] = qiskit_result[key]

            return result_dic  

        # Call VirtualBackend constructor
        if isinstance(backend.name, str):
            name = backend.name
        else:
            name = backend.name()

        super().__init__(run, port=port)


def VirtualQiskitBackend(*args, **kwargs):
    import warnings

    warnings.warn(
        "VirtualQiskitBackend will be deprecated in a future release of Qrisp. Use QiskitBackend instead."
    )
    return QiskitBackend(*args, **kwargs)


class QiskitRuntimeBackend(VirtualBackend):
    """
    This class instantiates a VirtualBackend using a Qiskit Runtime backend.
    This allows easy access to Qiskit Runtime backends through the qrisp interface.
    It is imporant to close the session after the execution of the algorithm
    (as reported in the example below).

    Parameters
    ----------
    api_token : str
        The token is necessary to create correctly the Qiskit Runtime
        service and be able to run algorithms on their backends.
    backend : str, optional
        A string associated to the name of a Qiskit Runtime backend.
        By default, the least busy available backend is selected.
    channel : str, optional
        The channel type. Available are ``ibm_cloud`` or ``ibm_quantum_platform``.
        The default is ``ibm_cloud``.
    mode : str, optional
        The `execution mode <https://quantum.cloud.ibm.com/docs/en/guides/execution-modes>`_. Available are ``job`` and ``session``.
        The default is ``job``. 

    Attributes
    ----------
    session : Session
        The `Qiskit Runtime session <https://quantum.cloud.ibm.com/docs/en/api/qiskit-ibm-runtime/session>`_.

    Examples
    --------

    >>> from qrisp import QuantumFloat
    >>> from qrisp.interface import QiskitRuntimeBackend
    >>> example_backend = QiskitRuntimeBackend(api_token = "YOUR_IBM_CLOUD_TOKEN", backend = "ibm_brisbane", channel = "ibm_cloud")
    >>> qf = QuantumFloat(2)
    >>> qf[:] = 2
    >>> res = qf*qf
    >>> result = res.get_measurement(backend = example_backend)
    >>> print(result)
    >>> # example_backend.close_session() # Use only when mode = "session"
    {4: 0.6133,
    8: 0.1126,
    0: 0.0838,
    12: 0.0798,
    5: 0.0272,
    6: 0.016,
    9: 0.0125,
    1: 0.0117,
    13: 0.0081,
    14: 0.0073,
    3: 0.0071,
    2: 0.0062,
    10: 0.0051,
    7: 0.0044,
    11: 0.0035,
    15: 0.0014}

    """

    def __init__(self, api_token, backend=None, channel="ibm_cloud", mode="job"):

        try:
            from qiskit_ibm_runtime import QiskitRuntimeService, SamplerV2, Session
        except ImportError:
            raise ImportError("Please install qiskit-ibm-runtime to use the QiskitBackend. You can do this by running `pip install qiskit-ibm-runtime`.")

        service = QiskitRuntimeService(channel=channel, token=api_token)
        if backend is None:
            backend = service.least_busy()
        else:
            backend = service.backend(backend)
            
        if mode == "session":
            self.session = Session(backend)
            sampler = SamplerV2(self.session)
        elif mode == "job":
            sampler = SamplerV2(backend)
        else:
            raise ValueError(f"Execution mode" + str(mode) + " not available.")

        # Create the run method
        def run(qasm_str, shots=None, token=""):
            if shots is None:
                shots = 1000
            # Convert to qiskit
            from qiskit import QuantumCircuit

            qiskit_qc = QuantumCircuit.from_qasm_str(qasm_str)

            # Make circuit with one monolithic register
            new_qiskit_qc = QuantumCircuit(len(qiskit_qc.qubits), len(qiskit_qc.clbits))
            for instr in qiskit_qc:
                new_qiskit_qc.append(
                    instr.operation,
                    [qiskit_qc.qubits.index(qb) for qb in instr.qubits],
                    [qiskit_qc.clbits.index(cb) for cb in instr.clbits],
                )

            from qiskit import transpile

            qiskit_qc = transpile(new_qiskit_qc, backend=backend)

            job = sampler.run([qiskit_qc], shots=shots)     

            qiskit_result = (
                job.result()[0].data.c.get_counts() 
                # https://docs.quantum.ibm.com/migration-guides/v2-primitives
            )

            # Remove the spaces in the qiskit result keys
            result_dic = {}
            import re

            for key in qiskit_result.keys():
                counts_string = re.sub(r"\W", "", key)
                result_dic[counts_string] = qiskit_result[key]

            return result_dic  

        super().__init__(run)

    def close_session(self):
        """
        Method to call in order to close the session started by the init method.
        """
        self.session.close()



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import time

import numpy as np
from scipy.optimize import minimize
from sympy import Symbol

from qrisp.algorithms.vqe.vqe_benchmark_data import VQEBenchmark
from qrisp.operators.qubit.measurement import QubitOperatorMeasurement
from qrisp.operators.fermionic import FermionicOperator

import jax
import jax.numpy as jnp
from qrisp.jasp import check_for_tracing_mode
from qrisp.jasp.optimization_tools.optimize import minimize as jasp_minimize


class VQEProblem:
    r"""
    Central structure to facilitate treatment of VQE problems.
    This class encapsulates the Hamiltonian, the ansatz, and the initial state preparation function for a specific VQE problem instance. 
    
    Parameters
    ----------
    hamiltonian : :ref:`QubitOperator` or :ref:`FermionicOperator`
        The problem Hamiltonian.
    ansatz_function : callable
        A function receiving a :ref:`QuantumVariable`, and an array of real parameters of size ``(n,)``. 
        This function implements the unitary corresponding to one layer of the ansatz.
    num_params : int
        The number of parameters $n$ per layer of the ansatz.
    init_function : callable, optional
        A function receiving a :ref:`QuantumVariable` and preparing the initial state from the $\ket{0}$ state.
        By default, the inital state is the $\ket{0}$ state.
    callback : bool, optional
        If ``True``, intermediate results are stored. The default is ``False``.

    Examples
    --------

    For a quick demonstration, we show how to calculate the ground state energy of the $H_2$ molecule using VQE. 
    The :meth:`electronic_structure_problem <qrisp.vqe.problems.electronic_structure.electronic_structure_problem>` method generates 
    a VQEProblem instance with the Hamiltonian and a chemistry-inspired Qubit Coupled Cluster Single Double `(QCCSD) ansatz <https://arxiv.org/abs/2005.08451>`_.

    ::

        from pyscf import gto
        from qrisp import QuantumVariable
        from qrisp.vqe.problems.electronic_structure import *

        mol = gto.M(
            atom = '''H 0 0 0; H 0 0 0.74''',
            basis = 'sto-3g')

        vqe = electronic_structure_problem(mol)

        energy = vqe.run(QuantumVariable(4), depth=1, max_iter=50)
        print(energy)
        #Yields -1.8461290172512965

    You can also specify a custom Hamiltonian and a custom hardware-efficient ansatz, as explained `here <https://arxiv.org/abs/2305.07092>`_.

    ::

        from qrisp import *
        from qrisp.operators.qubit import X,Y,Z

        # Problem Hamiltonian
        c = [-0.81054, 0.16614, 0.16892, 0.17218, -0.22573, 0.12091, 0.166145, 0.04523]
        H = c[0] \
            + c[1]*Z(0)*Z(2) \
            + c[2]*Z(1)*Z(3) \
            + c[3]*(Z(3) + Z(1)) \
            + c[4]*(Z(2) + Z(0)) \
            + c[5]*(Z(2)*Z(3) + Z(0)*Z(1)) \
            + c[6]*(Z(0)*Z(3) + Z(1)*Z(2)) \
            + c[7]*(Y(0)*Y(1)*Y(2)*Y(3) + X(0)*X(1)*Y(2)*Y(3) + Y(0)*Y(1)*X(2)*X(3) + X(0)*X(1)*X(2)*X(3))

        # Ansatz
        def ansatz(qv,theta):
            for i in range(4):
                ry(theta[i],qv[i])
            for i in range(3):
                cx(qv[i],qv[i+1])
            cx(qv[3],qv[0])

        from qrisp.vqe.vqe_problem import *

        vqe = VQEProblem(hamiltonian = H,
                         ansatz_function = ansatz,
                         num_params = 4,
                         callback = True)

        energy = vqe.run(QuantumVariable(4),
                      depth = 1,
                      max_iter = 50)
        print(energy)
        # Yields -1.864179046
    
    Note that for comparing to the results in the aforementioned paper, we have to add the nuclear repulsion energy $E_{\text{nuc}}=0.72$ to the calculated electronic energy $E_{\text{el}}$.

    We visualize the optimization process:

    >>> vqe.visualize_energy(exact=True)

    .. figure:: /_static/vqeH2.png
        :alt: VQEH2
        :scale: 80%
        :align: center

    """

    def __init__(
        self,
        hamiltonian,
        ansatz_function,
        num_params,
        init_function=None,
        callback=False,
    ):

        if isinstance(hamiltonian, FermionicOperator):
            hamiltonian = hamiltonian.to_qubit_operator()

        hamiltonian = hamiltonian.hermitize()
        hamiltonian = hamiltonian.eliminate_ladder_conjugates()
        hamiltonian = hamiltonian.apply_threshold(0)

        self.hamiltonian = hamiltonian
        self.ansatz_function = ansatz_function
        self.num_params = num_params
        self.init_function = init_function

        # parameters for callback
        self.callback = callback
        self.optimization_params = []
        self.optimization_costs = []

    def set_callback(self):
        """
        Sets ``callback=True`` for saving intermediate results.

        """

        self.callback = True

    def set_init_function(self, init_function):
        """
        Set the initial state preparation function for the VQE problem.

        Parameters
        ----------
        init_function : callable
            The initial state preparation function for the specific VQE problem instance.
        """
        self.init_function = init_function

    def compile_circuit(self, qarg, depth):
        """
        Compiles the circuit that is evaluated by the :meth:`run <qrisp.vqe.VQEProblem.run>` method.

        Parameters
        ----------
        qarg : :ref:`QuantumVariable`
            The argument to which the VQE circuit is applied.
        depth : int
            The amount of VQE layers.

        Returns
        -------
        compiled_qc : :ref:`QuantumCircuit`
            The parametrized, compiled quantum circuit without measurements.
        list[sympy.Symbol]
            A list of the parameters that appear in ``compiled_qc``.

        """

        temp = list(qarg.qs.data)

        # Define parameters theta for VQE circuit
        theta = [
            Symbol("theta_" + str(i) + str(j))
            for i in range(depth)
            for j in range(self.num_params)
        ]

        # Prepare the initial state for particular problem instance, the default is the \ket{0} state
        if self.init_function is not None:
            self.init_function(qarg)

        # Apply p layers of the ansatz
        for i in range(depth):
            self.ansatz_function(
                qarg, [theta[self.num_params * i + j] for j in range(self.num_params)]
            )

        # Compile quantum circuit
        compiled_qc = qarg.qs.compile()

        qarg.qs.data = temp

        return compiled_qc, theta

    def optimization_routine(
        self,
        qarg_prep,
        depth,
        mes_kwargs,
        init_type,
        init_point,
        optimizer,
        options,
        measurement_data,
    ):
        """
        Wrapper subroutine for the optimization method used in QAOA. The initial values are set and the optimization via ``COBYLA`` is conducted here.

        Parameters
        ----------
        qarg : :ref:`QuantumVariable` or callable
            The argument to which the VQE circuit is applied,
            or a function returning a :ref:`QuantumVariable` to which the VQE circuit is applied.
        depth : int
            The amount of VQE layers.
        mes_kwargs : dict
            The keyword arguments for the measurement function.
        init_type : string
            Specifies the way the initial optimization parameters are chosen.
        init_point : ndarray, shape (n,)
            Specifies the initial optimization parameters.
        optimizer : str
            Specifies the optimization routine.
        options : dict
            A dictionary of solver options.
        measurement_data : QubitOperatorMeasurement
            Cached data to accelerate the measurement procedure. Automatically generated by default.

        Returns
        -------
        ndarray, shape (n,)
            The optimized parameters of the problem instance.
        float or jax.Array
            The expectation value of the Hamiltonian for the optimized parameters.
        """

        if check_for_tracing_mode():

            # Define optimization wrapper function to be minimized using VQE
            def optimization_wrapper(theta, state_prep, mes_kwargs):
                """
                Wrapper function for the optimization method used in VQE.

                This function calculates the expectation value of the Hamiltonian.

                Parameters
                ----------
                theta : jax.Array
                    The array of angle parameters for the VQE circuit.
                state_prep : callable
                    A function returning a QuantumVariable.
                    The expectation of the Hamiltonian for the state of this QuantumVariable will be measured.
                    The state preparation function can only take classical values as arguments.
                    This is because a quantum value would need to be copied for each sampling iteration, which is prohibited by the no-cloning theorem.
                mes_kwargs : dict
                    The keyword arguments for the measurement function.

                Returns
                -------
                jax.Array
                    The expectation value of the Hamiltonian.
                """

                expectation = self.hamiltonian.expectation_value(
                    state_prep, **mes_kwargs
                )(theta)

                return expectation

        else:

            # Define optimization wrapper function to be minimized using VQE
            def optimization_wrapper(
                theta, state_prep, mes_kwargs, compiled_qc, symbols, measurement_data
            ):
                """
                Wrapper function for the optimization method used in VQE.

                This function calculates the expectation value of the Hamiltonian.

                Parameters
                ----------
                theta : list
                    The list of angle parameters for the VQE circuit.
                state_prep : callable
                    A function returning a QuantumVariable.
                mes_kwargs : dict
                    The keyword arguments for the measurement function.
                compiled_qc : QuantumCircuit
                    The compiled quantum circuit.
                symbols : list
                    The list of symbols used in the quantum circuit.
                measurement_data : QubitOperatorMeasurement
                    Cached data to accelerate the measurement procedure. Automatically generated by default.

                Returns
                -------
                float
                    The expectation value of the Hamiltonian.
                """

                subs_dic = {symbols[i]: theta[i] for i in range(len(symbols))}

                expectation = self.hamiltonian.expectation_value(
                    state_prep,
                    **mes_kwargs,
                    precompiled_qc=compiled_qc,
                    subs_dic=subs_dic,
                    measurement_data=measurement_data,
                )(theta)

                if self.callback:
                    self.optimization_costs.append(expectation)

                return expectation

        # Initialization for optimization parameters
        if init_point is None:

            if init_type == "random":
                # Random optimization parameters
                if check_for_tracing_mode():
                    key = jax.random.key(11)
                    init_point = (
                        jax.random.uniform(key=key, shape=(self.num_params * depth,))
                        * jnp.pi
                        / 2
                    )
                else:
                    init_point = np.random.rand(self.num_params * depth) * np.pi / 2

            else:
                raise Exception(
                    f"Parameter initialization method {init_type} is not available."
                )

        def state_prep(theta):

            qarg = qarg_prep()

            # Prepare the initial state, the default is the \ket{0} state
            if self.init_function is not None:
                self.init_function(qarg)

            for i in range(depth):
                self.ansatz_function(
                    qarg, theta[self.num_params * i : self.num_params * (i + 1) + 1]
                )

            return qarg

        # Perform optimization using specified method
        if check_for_tracing_mode():

            res_sample = jasp_minimize(
                optimization_wrapper,
                init_point,
                method=optimizer,
                options=options,
                args=(
                    state_prep,
                    mes_kwargs,
                ),
            )

            return res_sample.x, res_sample.fun

        else:

            compiled_qc, symbols = self.compile_circuit(qarg_prep(), depth)

            res_sample = minimize(
                optimization_wrapper,
                init_point,
                method=optimizer,
                options=options,
                args=(
                    state_prep,
                    mes_kwargs,
                    compiled_qc,
                    symbols,
                    measurement_data,
                ),
            )

            return res_sample.x, res_sample.fun

    def run(
        self,
        qarg,
        depth,
        mes_kwargs={},
        max_iter=50,
        init_type="random",
        init_point=None,
        optimizer="COBYLA",
        options={},
    ):
        """
        Run VQE for the specific problem instance.

        Parameters
        ----------
        qarg : :ref:`QuantumVariable` or callable
            The argument to which the VQE circuit is applied,
            or a function returning a :ref:`QuantumVariable` to which the VQE circuit is applied.
        depth : int
            The amount of VQE ansatz layers.
        mes_kwargs : dict, optional
            The keyword arguments for the :meth:`expectation_value <qrisp.operators.qubit.QubitOperator.expectation_value>` function. Default is an empty dictionary.
            By default, the target ``precision`` is set to 0.01. Precision refers to how accurately the Hamiltonian is evaluated.
            The number of shots the backend performs per iteration scales quadratically with the inverse precision.
        max_iter : int, optional
            The maximum number of iterations for the optimization method. Default is 50.
        init_type : string, optional
            Specifies the way the initial optimization parameters are chosen. Available is ``random``.
            The default is ``random``: Parameters are initialized uniformly at random in the interval $[0,\pi/2)]$.
        init_point : ndarray, shape (n,), optional
            Specifies the initial optimization parameters.
        optimizer : str, optional
            Specifies the `SciPy optimization routine <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html>`_.
            Available are, e.g., ``COBYLA``, ``COBYQA``, ``Nelder-Mead``. The Default is ``COBYLA``.
            In tracing mode (i.e. Jasp) Jax-traceable :ref:`optimization routines <optimization_tools>` must be utilized.
            Available are ``COBYLA``, ``SPSA``.
        options : dict
            A dictionary of solver options.

        Returns
        -------
        float or jax.Array
            The expectation value of the Hamiltonian after applying the optimal VQE circuit to the quantum argument.

        """

        if callable(qarg):
            qarg_prep = qarg
        else:
            template = qarg.template()

            def qarg_prep():
                return template.construct()

        # Set default options
        if not "precision" in mes_kwargs:
            mes_kwargs["precision"] = 0.01

        if not "diagonalisation_method" in mes_kwargs:
            mes_kwargs["diagonalisation_method"] = "commuting_qw"

        options["maxiter"] = max_iter

        if check_for_tracing_mode():

            measurement_data = None

        else:

            measurement_data = QubitOperatorMeasurement(
                self.hamiltonian,
                diagonalisation_method=mes_kwargs["diagonalisation_method"],
            )

        opt_theta, opt_res = self.optimization_routine(
            qarg_prep,
            depth,
            mes_kwargs,
            init_type,
            init_point,
            optimizer,
            options,
            measurement_data=measurement_data,
        )

        return opt_res

    def train_function(
        self,
        qarg,
        depth,
        mes_kwargs={},
        max_iter=50,
        init_type="random",
        init_point=None,
        optimizer="COBYLA",
        options={},
    ):
        """
        This function allows for training of a circuit with a given instance of a ``VQEProblem``. It will then return a function that can be applied to a :ref:`QuantumVariable`,
        such that it prepares the ground state of the problem Hamiltonian. The function therefore applies a circuit for the problem instance with optimized parameters.

        Parameters
        ----------
        qarg : :ref:`QuantumVariable` or callable
            The argument to which the VQE circuit is applied,
            or a function returning a :ref:`QuantumVariable` to which the VQE circuit is applied.
        depth : int
            The amount of VQE ansatz layers.
        mes_kwargs : dict, optional
            The keyword arguments for the :meth:`expectation_value <qrisp.operators.qubit.QubitOperator.expectation_value>` function. Default is an empty dictionary.
            By default, the target ``precision`` is set to 0.01. Precision refers to how accurately the Hamiltonian is evaluated.
            The number of shots the backend performs per iteration scales quadratically with the inverse precision.
        max_iter : int, optional
            The maximum number of iterations for the optimization method. Default is 50.
        init_type : string, optional
            Specifies the way the initial optimization parameters are chosen. Available is ``random``.
            The default is ``random``: Parameters are initialized uniformly at random in the interval $[0,\pi/2)]$.
        init_point : ndarray, shape (n,), optional
            Specifies the initial optimization parameters.
        optimizer : str, optional
            Specifies the `SciPy optimization routine <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html>`_.
            Available are, e.g., ``COBYLA``, ``COBYQA``, ``Nelder-Mead``. The Default is ``COBYLA``.
            In tracing mode (i.e. Jasp) Jax-traceable :ref:`optimization routines <optimization_tools>` must be utilized.
            Available are ``COBYLA``, ``SPSA``.
        options : dict
            A dictionary of solver options.

        Returns
        -------
        callable
            A function that can be applied to a :ref:`QuantumVariable`, with optimized parameters for the problem instance.
            The :ref:`QuantumVariable` then represents the ground state of the problem Hamiltonian.

        """

        if callable(qarg):
            qarg_prep = qarg
        else:
            template = qarg.template()

            def qarg_prep():
                return template.construct()

        # Set default options
        if not "precision" in mes_kwargs:
            mes_kwargs["precision"] = 0.01

        if not "diagonalisation_method" in mes_kwargs:
            mes_kwargs["diagonalisation_method"] = "commuting_qw"

        options["maxiter"] = max_iter

        if check_for_tracing_mode():

            measurement_data = None

        else:

            measurement_data = QubitOperatorMeasurement(
                self.hamiltonian,
                diagonalisation_method=mes_kwargs["diagonalisation_method"],
            )

        opt_theta, opt_res = self.optimization_routine(
            qarg_prep,
            depth,
            mes_kwargs,
            init_type,
            init_point,
            optimizer,
            options,
            measurement_data=measurement_data,
        )

        def circuit_generator(qarg):

            if self.init_function is not None:
                self.init_function(qarg)

            for i in range(depth):
                self.ansatz_function(
                    qarg, opt_theta[self.num_params * i : self.num_params * (i + 1) + 1]
                )

        return circuit_generator

    def benchmark(
        self,
        qarg,
        depth_range,
        precision_range,
        iter_range,
        optimal_energy,
        repetitions=1,
        mes_kwargs={},
        init_type="random",
        optimizer="COBYLA",
        options={},
    ):
        """
        This method enables convenient data collection regarding performance of the implementation.

        Parameters
        ----------
        qarg : :ref:`QuantumVariable` or callable
            The argument to which the VQE circuit is applied,
            or a function returning a :ref:`QuantumVariable` to which the VQE circuit is applied.
        depth_range : list[int]
            A list of integers indicating, which depth parameters should be explored. Depth means the amount of VQE ansatz layers.
        precision_range : list[float]
            A list of floats indicating, which precision parameters should be explored. Precision refers to how accurately the Hamiltonian is evaluated.
            The number of shots the backend performs per iteration scales quadratically with the inverse precision.
        iter_range : list[int]
            A list of integers indicating, what iterations parameter should be explored. Iterations means the amount of backend calls, the optimizer is allowed to do.
        optimal_energy : float
            The exact ground state energy of the problem Hamiltonian.
        repetitions : int, optional
            The amount of repetitions, each parameter constellation should go though. Can be used to get a better statistical significance. The default is 1.
        mes_kwargs : dict, optional
            The keyword arguments for the :meth:`expectation_value <qrisp.operators.qubit.QubitOperator.expectation_value>` function. Default is an empty dictionary.
        init_type : str, optional
            Specifies the way the initial optimization parameters are chosen. Available is ``random``.
            The default is ``random``: Parameters are initialized uniformly at random in the interval $[0,\pi/2)]$.
        optimizer : str, optional
            Specifies the `SciPy optimization routine <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html>`_.
            Available are, e.g., ``COBYLA``, ``COBYQA``, ``Nelder-Mead``. The Default is ``COBYLA``.
        options : dict
            A dictionary of solver options.

        Returns
        -------
        :ref:`VQEBenchmark`
            The results of the benchmark.

        Examples
        --------

        We create a Heisenberg problem instance and benchmark several parameters:

        ::

            from qrisp import QuantumVariable
            from qrisp.vqe.problems.heisenberg import *
            from networkx import Graph

            G = Graph()
            G.add_edges_from([(0,1),(1,2),(2,3),(3,4)])

            vqe = heisenberg_problem(G,1,0)
            H = create_heisenberg_hamiltonian(G,1,0)

            benchmark_data = vqe.benchmark(QuantumVariable(5),
                                depth_range = [1,2,3],
                                precision_range = [0.02,0.01],
                                iter_range = [25,50],
                                optimal_energy = H.ground_state_energy(),
                                repetitions = 2
                                )

        We can investigate the data by calling ``visualize``:

        ::

            benchmark_data.visualize()

        .. image:: vqe_benchmark_plot.png

        The :ref:`VQEBenchmark` class contains a variety of methods to help
        you drawing conclusions from the collected data. Make sure to check them out!

        """

        if callable(qarg):
            qarg_prep = qarg
        else:
            template = qarg.template()

            def qarg_prep():
                return template.construct()

        data_dict = {
            "layer_depth": [],
            "circuit_depth": [],
            "qubit_amount": [],
            "precision": [],
            "iterations": [],
            "runtime": [],
            "energy": [],
        }

        for p in depth_range:
            for s in precision_range:
                for it in iter_range:
                    for k in range(repetitions):

                        start_time = time.time()

                        temp_mes_kwargs = dict(mes_kwargs)
                        temp_mes_kwargs["precision"] = s

                        energy = self.run(
                            qarg_prep,
                            depth=p,
                            max_iter=it,
                            mes_kwargs=temp_mes_kwargs,
                            init_type=init_type,
                            optimizer=optimizer,
                            options=options,
                        )

                        final_time = time.time() - start_time

                        compiled_qc, _ = self.compile_circuit(qarg_prep(), depth=p)

                        data_dict["layer_depth"].append(p)
                        data_dict["circuit_depth"].append(compiled_qc.depth())
                        data_dict["qubit_amount"].append(compiled_qc.num_qubits())
                        data_dict["precision"].append(s)
                        data_dict["iterations"].append(it)
                        data_dict["energy"].append(energy)
                        data_dict["runtime"].append(final_time)

        return VQEBenchmark(data_dict, optimal_energy, self.hamiltonian)

    def visualize_energy(self, exact=False):
        """
        Visualizes the energy during the optimization process.

        Parameters
        ----------
        exact : Boolean
            If ``True``, the exact ground state energy of the spin operator is computed classically, and compared to the energy in the optimization process.
            The default is ``False``.

        """

        import matplotlib.pyplot as plt

        if not self.callback:
            raise Exception(
                "Visualization can only be performed for a VQE instance with callback=True"
            )

        if exact:
            exact_solution = self.hamiltonian.ground_state_energy()
            plt.axhline(
                y=exact_solution,
                color="#6929C4",
                linestyle="--",
                linewidth=2,
                label="Exact energy",
            )

        x = list(range(len(self.optimization_costs)))
        y = self.optimization_costs
        plt.scatter(
            x,
            y,
            color="#20306f",
            marker="o",
            linestyle="solid",
            linewidth=1,
            label="VQE energy",
        )
        plt.xlabel("Iterations", fontsize=15, color="#444444")
        plt.ylabel("Energy", fontsize=15, color="#444444")
        plt.legend(fontsize=12, labelcolor="#444444")
        plt.tick_params(axis="both", labelsize=12)
        plt.grid()

        plt.show()



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from typing import Union

from qrisp.alg_primitives.arithmetic.adders.gidney import gidney_adder
from qrisp.qtypes import QuantumFloat, QuantumModulus
from qrisp.jasp import jrange, check_for_tracing_mode, jlen, q_cond
from qrisp.environments import control, invert, custom_control
from qrisp.core import swap, cx, x

from .jasp_bigintiger import BigInteger
from .jasp_mod_tools import montgomery_encoder, montgomery_decoder, modinv, best_montgomery_shift


def q_montgomery_reduction(qf: QuantumFloat, N: Union[int, BigInteger], m: int, inpl_adder=gidney_adder):
    """
    Perform the Montgomery reduction of a concatenated QuantumFloat in-place.

    Layout
    ------
    qf = aux[:] + res[:]
    - aux has m+1 qubits (for u-tilde including the folded sign bit),
    - res has n qubits (holds the reduced result).

    Algorithm
    ---------
    - Estimation (m steps): For each LSB, conditionally subtract floor(N/2) on the
      truncated slice (implicit right-shift).
    - Correction: If negative, add N to the result part; fold the sign bit into
      aux via a single CNOT to produce u-tilde (m+1 bits).

    Parameters
    ----------
    qf : QuantumFloat
        The concatenated QuantumFloat (aux || res).
    N : int or BigInteger
        Modulus (must be odd).
    m : int
        Exponent m of the auxiliary radix R = 2^m.
    inpl_adder : Callable
        In-place adder to use during computation (defaults to gidney_adder).
    """
    if check_for_tracing_mode():
        xrange = jrange
    else:
        xrange = range

    # Estimation stage
    for j in xrange(m):
        with control(qf[j]):
            with invert():
                inpl_adder((N - 1) >> 1, qf[j + 1:])

    # Correction stage
    with control(qf[-1]):
        inpl_adder(N, qf[m:-1])

    # Move sign bit across the result into the aux MSB (n swaps)
    for i in xrange(jlen(qf) - m - 1):
        swap(qf[-1 - i], qf[-2 - i])

    # Fold sign with LSB of estimate to form u-tilde
    cx(qf[m + 1], qf[m])


def cq_montgomery_multiply(X: Union[int, BigInteger], y: QuantumFloat, N: Union[int, BigInteger], m: int,
                           inpl_adder=gidney_adder, x_is_montgomery: bool = False, res=None):
    """
    Montgomery product of a classical X and a QuantumFloat y: X*y*R^{-1} mod N.

    Outline
    -------
    - If needed, Montgomery-encode X using R = 2^m.
    - Accumulate reduced partial products (X*2^j mod N) into wqf = aux || res.
    - Apply q_montgomery_reduction(wqf, N, m).
    - Uncompute aux (size m+1) by adding ((X*2^j mod N) * N^{-1} mod 2^{m+1})
      controlled by y_j, with invert().

    Important
    ---------
    In the quantum-classical case (Rines & Chuang, 2018), the quantum input y
    can remain in standard (non-Montgomery) representation, and the output is
    returned in standard representation:
      output = X*y mod N

    Parameters
    ----------
    X : int or BigInteger
        Classical multiplicand. If not already in Montgomery form, set
        x_is_montgomery=False (default), which will encode it.
    y : QuantumFloat
        Quantum multiplicand (n qubits, standard representation).
    N : int or BigInteger
        Odd modulus.
    m : int
        Exponent m of the auxiliary radix R = 2^m.
    inpl_adder : Callable
        In-place adder to use during computation (defaults to gidney_adder).
    x_is_montgomery : bool
        If the classical X is already in Montgomery form. Defaults to False.
    res : QuantumFloat or None
        Optional target result register (n qubits). Allocated if None.

    Returns
    -------
    QuantumFloat
        The Montgomery product X*y mod N in standard representation.
    """
    # Build R = 2^m with width matching X if BigInteger
    if isinstance(X, BigInteger):
        R = BigInteger.create(1, X.digits.shape[0]) << m
    else:
        R = 1 << m

    if not x_is_montgomery:
        X = montgomery_encoder(X, R, N)

    # N^{-1} modulo 2^{m+1}
    N1 = modinv(N, R << 1)

    if check_for_tracing_mode():
        xrange = jrange
    else:
        xrange = range
    n = jlen(y)
    if res is None:
        res = QuantumFloat(n)
    aux = QuantumFloat(m + 1)
    wqf = aux[:] + res[:]

    # Multiplication: sum_j y_j * (X*2^j mod N) into wqf
    for i in xrange(n):
        j = n - i - 1
        with control(y[j]):
            inpl_adder((X << j) % N, wqf)

    # Reduction
    q_montgomery_reduction(wqf, N, m, inpl_adder=inpl_adder)

    # Uncompute aux: add ((X*2^j mod N)*N^{-1} mod 2^{m+1})
    for i in xrange(n):
        j = n - i - 1
        with control(y[j]):
            with invert():
                inpl_adder(((X << j) % N) * N1, aux[:])

    aux.delete()
    return res


@custom_control
def cq_montgomery_multiply_inplace(X: Union[int, BigInteger], y: QuantumFloat, N: Union[int, BigInteger], m: int,
                                   inpl_adder=gidney_adder, x_is_montgomery: bool = False, ctrl=None):
    """
    Montgomery product of a classical X and a QuantumFloat y, in-place on y.

    Notes
    -----
    - y remains in standard representation; the output is standard (X*y mod N).

    Parameters
    ----------
    X : int or BigInteger
        Integer factor of the Montgomery product.
    y : QuantumFloat
        Quantum factor of the Montgomery product (overwritten in-place).
    N : int or BigInteger
        Modulus of the Montgomery reduction.
    m : int
        Exponent m of the auxiliary radix R = 2^m.
    inpl_adder : Callable
        In-place adder to use during computation.
    x_is_montgomery : bool
        If the classical input X is already in Montgomery form. Defaults to False.
    ctrl : QuantumBit or None
        Optional external control for the in-place operation.
    """
    tmp = QuantumFloat(y.size)

    if ctrl is not None:
        x(ctrl)
        with control(ctrl):  # TODO: Add invert=True
            for i in jrange(y.size):
                swap(tmp[i], y[i])
        x(ctrl)

    cq_montgomery_multiply(X, y, N, m, inpl_adder, x_is_montgomery, tmp)

    if x_is_montgomery:
        if isinstance(X, BigInteger):
            R = BigInteger.create(1, X.digits.shape[0]) << m
            X = montgomery_decoder(X, R, N)
        else:
            X = montgomery_decoder(X, 1 << m, N)
    X1 = modinv(X, N)

    if ctrl is not None:
        x(ctrl)
        with control(ctrl):  # TODO: Add invert=True
            for i in jrange(y.size):
                swap(tmp[i], y[i])
        x(ctrl)

    with invert():
        cq_montgomery_multiply(X1, tmp, N, m, inpl_adder, False, y)

    if ctrl is not None:
        with control(ctrl, invert=False):
            for i in jrange(y.size):
                swap(tmp[i], y[i])
    else:
        for i in jrange(y.size):
            swap(tmp[i], y[i])

    tmp.delete()


def qq_montgomery_multiply(x: QuantumFloat, y: QuantumFloat, N: int, m: int, inpl_adder=gidney_adder):
    """
    Perform the montgomery product of two QuantumFloats. Note that both QuantumFloats must be in montgomery form.

    Parameters
    ----------
    x : QuantumFloat
        First factor of the montgomery product.
    y : QuantumFloat
        Second factor of the montgomery product.
    N : int
        Original modulus of the montgomery reduction.
    m : int
        Exponent $m$ of the auxiliary radix $R=2^m$ (see best_montgomery_shift).
    inpl_adder : Callable
        In-place adder to use during computation

    Returns
    ----------
    QuantumFloat
        The mongomery product of the inputs.
    """

    if check_for_tracing_mode():
        xrange = jrange
    else:
        xrange = range

    def qq_mul(ox: QuantumFloat, oy: QuantumFloat, ores: QuantumFloat):
        for i in xrange(jlen(oy)):
            with control(oy[i]):
                inpl_adder(ox[:], ores[i:])

    def qc_mul_inplace(operand, cl_int):
        size = jlen(operand)
        for i in xrange(size - 1):
            with control(operand[size - 2 - i]):
                inpl_adder(cl_int // 2, operand[size - 1 - i:])

    n = jlen(y)
    res = QuantumFloat(n)
    aux = QuantumFloat(m + 1)
    wqf = aux[:] + res[:]

    qq_mul(x, y, wqf[:-1])
    q_montgomery_reduction(wqf, N, m, inpl_adder=inpl_adder)
    qc_mul_inplace(aux, N)
    with invert():
        qq_mul(x, y, aux[:])
    aux.delete()

    return res


def qq_montgomery_multiply_modulus(x: QuantumModulus, y: QuantumModulus):
    """
    Perform the montgomery product of two QuantumModuli. Note that both QuantumModuli must be in montgomery form.
    Similiar to `qq_montgomery_multiply` but extracts the fields of the Moduli.
    Assumes that both QuantumModuli share the same modulus and the montgomery shift.

    Parameters
    ----------
    x : QuantumModulus
        First factor of the montgomery product.
    y : QuantumModulus
        Second factor of the montgomery product.

    Returns
    ----------
    QuantumModulus
        The mongomery product of the inputs.
    """

    inpl_adder = x.inpl_adder
    N = x.modulus
    m = x.m

    # res = qq_montgomery_multiply(x, y, N, m, inpl_adder)
    # return res

    if check_for_tracing_mode():
        xrange = jrange
    else:
        xrange = range

    def qq_mul(ox: QuantumFloat, oy: QuantumFloat, ores: QuantumFloat):
        for i in xrange(jlen(oy)):
            with control(oy[i]):
                inpl_adder(ox[:], ores[i:])

    def qc_mul_inplace(operand, cl_int):
        size = jlen(operand)
        for i in xrange(size - 1):
            with control(operand[size - 2 - i]):
                inpl_adder(cl_int // 2, operand[size - 1 - i:])

    res = QuantumModulus(N)
    res.m = m
    aux = QuantumFloat(m + 1)
    wqf = aux[:] + res[:]

    qq_mul(x, y, wqf[:-1])
    q_montgomery_reduction(wqf, N, m, inpl_adder=inpl_adder)
    qc_mul_inplace(aux, N)
    with invert():
        qq_mul(x, y, aux[:])
    aux.delete()

    return res

def cq_montgomery_mat_multiply(A, B, out):
    if check_for_tracing_mode():
        xrange = jrange
        x_cond = q_cond
    else:
        xrange = range
        def x_cond(condition, tf, ff):
            if condition:
                tf()
            else:
                ff()

    n1 = A.shape[0]
    n2 = B.shape[1]

    m = A.shape[1]

    #out = QuantumArray(qtype=A[0,0], shape=(n1, n2))

    for k in xrange(m):
        for j in xrange(n2):
            for i in xrange(n1):
                def true_fun():
                    best_montgomery_shift(B[k,j], A[i,k].modulus)
                    shift = best_montgomery_shift(B[k,j], A[i,k].modulus)
                    aux = cq_montgomery_multiply(B[k,j], A[i,k], A[i,k].modulus, shift)
                    out[i,j] += aux
                    with invert():
                        cq_montgomery_multiply(B[k,j], A[i,k], A[i,k].modulus, shift, res=aux)
                    aux.delete()
                x_cond(B[k,j] != 0, true_fun, lambda:None)

    return out


"""
********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from qrisp import z, control
from qrisp.alg_primitives.qae import amplitude_amplification
from qrisp.jasp import check_for_tracing_mode, expectation_value
from jax.lax import while_loop


def IQAE(qargs, state_function, eps, alpha, mes_kwargs={}):
    r"""
    Accelerated Quantum Amplitude Estimation (IQAE). This function performs :ref:`QAE <QAE>` with a fraction of the quantum resources of the well-known `QAE algorithm <https://arxiv.org/abs/quant-ph/0005055>`_.
    See `Accelerated Quantum Amplitude Estimation without QFT <https://arxiv.org/abs/2407.16795>`_.

    The problem of iterative quantum amplitude estimation is described as follows:

    * Given a unitary operator :math:`\mathcal{A}`, let :math:`\ket{\Psi}=\mathcal{A}\ket{0}\ket{\text{False}}`.
    * Write :math:`\ket{\Psi}=\sqrt{a}\ket{\Psi_1}\ket{\text{True}}+\sqrt{1-a}\ket{\Psi_0}\ket{\text{False}}` as a superposition of the orthogonal good and bad components of :math:`\ket{\Psi}`.
    * Find an estimate for $a$, the probability that a measurement of $\ket{\Psi}$ yields a good state.

    Parameters
    ----------
    qargs : list[:ref:`QuantumVariable`] or callable
        A list of QuantumVariables which represent the state on which the quantum amplitude estimation is performed,
        or a function preparing a list of QuantumVariables.
        The last variable in the list must be of type :ref:`QuantumBool`.
    state_function : callable
        A Python function preparing the state :math:`\ket{\Psi}`.
        This function will receive the variables returned by ``init_function`` as arguments.
    eps : float
        Accuracy $\epsilon>0$ of the algorithm.
    alpha : float
        Confidence level $\alpha\in (0,1)$ of the algorithm.
    mes_kwargs : dict, optional
        The keyword arguments for the measurement function. Default is an empty dictionary.

    Returns
    -------
    a : float
        An estimate $\hat{a}$ of $a$ such that

    .. math::

        \mathbb P\{|\hat{a}-a|<\epsilon\}\geq 1-\alpha

    Examples
    --------

    We show the same **Numerical integration** example which can also be found in the :ref:`QAE documentation <QAE>`.

    We wish to evaluate

    .. math::

        A=\int_0^1f(x)\mathrm dx.

    For this, we set up the corresponding ``state_function`` acting on the variables in ``input_list``:

    ::

        from qrisp import QuantumFloat, QuantumBool, control, z, h, ry, IQAE
        import numpy as np

        n = 6
        inp = QuantumFloat(n,-n)
        tar = QuantumBool()
        input_list = [inp, tar]

    For example, if $f(x)=\sin^2(x)$, the ``state_function`` can be implemented as follows:

    ::

        def state_function(inp, tar):
            h(inp)

            N = 2**inp.size
            for k in range(inp.size):
                with control(inp[k]):
                    ry(2**(k+1)/N,tar)

    Finally, we apply IQAE and obtain an estimate $a$ for the value of the integral $A=0.27268$.

    ::

        a = IQAE(input_list, state_function, eps=0.01, alpha=0.01)

    >>> a
    0.26782038552705856

    """

    if callable(qargs):

        init_function = qargs

    else:

        templates = [qv.template() for qv in qargs]

        def init_function():
            qargs_ = [temp.construct() for temp in templates]
            return qargs_

    # The oracle tagging the good states
    def oracle_function(*args):
        tar = args[-1]
        z(tar)

    if check_for_tracing_mode:
        import jax.numpy as jnp
    else:
        import numpy as jnp

    E = 1 / 2 * jnp.pow(jnp.sin(jnp.pi * 3 / 14), 2) - 1 / 2 * pow(
        jnp.sin(jnp.pi * 1 / 6), 2
    )
    F = 1 / 2 * jnp.arcsin(jnp.sqrt(2 * E))

    C = 4 / (6 * F + jnp.pi)
    break_cond = 2 * eps + 1

    K_i = 1
    m_i = 0

    theta_b = 0
    theta_sh = 0

    L_arr = jnp.array([3, 3, 3, 5, 5, 5, 5, 5, 7, 7, 7, 7, 7, 7, 7])
    m_arr = jnp.array([0, 1, 2, 0, 1, 2, 3, 4, 0, 1, 2, 3, 4, 5, 6])

    def cond_fun(state):
        L_arr, m_arr, break_cond, alpha, eps, m_i, K_i, theta_b, theta_sh = state
        return break_cond > 2 * eps

    def body_fun(state):
        L_arr, m_arr, break_cond, alpha, eps, m_i, K_i, theta_b, theta_sh = state

        alp_i = C * alpha * eps * K_i
        N_i = jnp.int64(jnp.ceil(1 / (2 * jnp.pow(E, 2)) * jnp.log(2 / alp_i)))

        # Perform quantum step
        A_i = quantum_step(
            jnp.int64((K_i - 1) / 2),
            N_i,
            init_function,
            state_function,
            oracle_function,
            mes_kwargs,
        )

        # Compute new thetas
        theta_b, theta_sh = compute_thetas(m_i, K_i, A_i, E)

        # Compute new L_i
        L_new, m_new = compute_Li(L_arr, m_arr, m_i, K_i, theta_b, theta_sh)
        m_i = m_new
        K_i = L_new * K_i

        break_cond = jnp.float64(jnp.abs(theta_b - theta_sh))

        return L_arr, m_arr, break_cond, alpha, eps, m_i, K_i, theta_b, theta_sh

    state = (L_arr, m_arr, break_cond, alpha, eps, m_i, K_i, theta_b, theta_sh)

    if check_for_tracing_mode():
        L_arr, m_arr, break_cond, alpha, eps, m_i, K_i, theta_b, theta_sh = while_loop(
            cond_fun, body_fun, state
        )
    else:
        while cond_fun(state):
            state = body_fun(state)
        L_arr, m_arr, break_cond, alpha, eps, m_i, K_i, theta_b, theta_sh = state

    final_res = jnp.sin((theta_b + theta_sh) / 2) ** 2
    return final_res


def quantum_step(k, N, init_function, state_function, oracle_function, mes_kwargs):
    """
    Performs the quantum step, i.e., Quantum Amplitude Amplification,
    in accordance to `Accelerated Quantum Amplitude Estimation without QFT <https://arxiv.org/abs/2407.16795>`_

    Parameters
    ----------
    k : int
        The amount of amplification steps, i.e., the power of :math:`\mathcal{Q}` in amplitude amplification.
    N : int
        The amount of shots, i.e., the amount of times the last qubit is measured after the amplitude amplification steps.
    init_function : callable
        A Python function that returns a list of QuantumVariables representing the state on which the quantum amplitude estimation is performed.
        The last variable in the list must be of type :ref:`QuantumBool`.
    state_function : callable
        A Python function preparing the state :math:`\ket{\Psi}`.
        This function will receive the variables in the list returnded by ``init_function`` as arguments.
    oracle_function : callable
        A Python function tagging the good state :math:`\ket{\Psi_1}`.
        This function will receive the variables in the list ``args`` as arguments in the
        course of this algorithm.
    mes_kwargs : dict, optional
        The keyword arguments for the measurement function. Default is an empty dictionary.
    """

    def state_prep(k):
        qargs = init_function()
        state_function(*qargs)
        amplitude_amplification(qargs, state_function, oracle_function, iter=k)
        return qargs[-1]

    if check_for_tracing_mode():
        a_i = expectation_value(state_prep, shots=N)(k)
    else:
        mes_kwargs["shots"] = N
        res_dict = state_prep(k).get_measurement(**mes_kwargs)
        a_i = res_dict.get(True, 0)

    return a_i


def compute_thetas(m_i, K_i, A_i, E):
    """
    Helper function to compute the angles for the next iteration.
    See `the original paper <https://arxiv.org/abs/2407.16795>`_ , Algorithm 1.

    Parameters
    ----------
    m_i : int
        Used for the computation of the interval of allowed angles.
    K_i : int
        Maximal amount of amplitude amplification steps for the next iteration.
    A_i : float
        Share of ``1``-measurements in amplitude amplification steps.
    E : float
        :math:`\epsilon` limit.
    """

    if check_for_tracing_mode:
        import jax.numpy as jnp
    else:
        import numpy as jnp

    b_max = jnp.max(jnp.array([A_i - E, 0]))
    sh_min = jnp.min(jnp.array([A_i + E, 1]))

    theta_b = (
        (m_i + m_i % 2) * jnp.pi / 2
        + jnp.pow(-1, m_i % 2) * jnp.arcsin(jnp.sqrt(b_max))
    ) / K_i
    theta_sh = (
        (m_i + m_i % 2) * jnp.pi / 2
        + jnp.pow(-1, m_i % 2) * jnp.arcsin(jnp.sqrt(sh_min))
    ) / K_i

    # assert np.round( np.pow( np.sin(K_i * theta_b),2) , 8 )  == np.round(b_max, 8)
    # assert np.round( np.pow( np.sin(K_i * theta_sh),2), 8 )  == np.round(sh_min, 8)

    return theta_b, theta_sh


def compute_Li(L_arr, m_arr, m_i, K_i, theta_b, theta_sh):
    """
    Helper function to compute further values for the next iteration.
    See `the original paper <https://arxiv.org/abs/2407.16795>`_ , Algorithm 1.

    Parameters
    ----------
    m_i : int
        Used for the computation of the interval of allowed angles.
    K_i : int
        Maximal amount of amplitude amplification steps for the next iteration.
    theta_b : float
        Lower bound for angle from last iteration.
    theta_b : float
        Upper bound for angle from last iteration.
    """

    if check_for_tracing_mode:
        import jax.numpy as jnp
    else:
        import numpy as jnp

    first_arr = L_arr * K_i * theta_b
    second_arr = L_arr * K_i * theta_sh

    lower_arr = (L_arr * m_i + m_arr) * jnp.pi / 2
    upper_arr = lower_arr + jnp.pi / 2

    index = jnp.argmax(
        (first_arr >= lower_arr)
        & (first_arr <= upper_arr)
        & (second_arr >= lower_arr)
        & (second_arr <= upper_arr)
    )

    L_new = L_arr[index]
    m_new = L_new * m_i + m_arr[index]

    return L_new, m_new



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import jax
import jax.numpy as jnp

from qrisp.jasp.tracing_logic import quantum_kernel, check_for_tracing_mode

# The following function implements the sample feature.

# The basic functionality would be relatively straightforward to implement,
# however there are some complications. The reason for that is that the resulting
# jaxpr should be "readable" by the terminal sampling interpreter.
# Terminal sampling means that instead of performing the simulations "shots"-times
# it is performed once and the shots are then sampled from that distribution.
# Naturally this implies a massive performance increase, which is why a lot
# of effort is spent to realize a smooth implementation.

# The underlying idea to make the feature easily "readable" by the terminal
# sampling interpreter is to structure one iteration of sampling into three
# steps.

# 1. Evaluating the user function, which generates the distribution.
# 2. Sampling from that distribution via the "measure" function.
# 3. Decoding and postprocessing the measurement results.

# For the final two steps we deploy some custom logic to realize the terminal
# sampling behavior. To simplify the automatic processing of these steps,
# we capture each into individual pjit calls.

# The terminal sampling interpreter then identifies each steps via the
# eqn.params["name"] attribute and executes the custom logic.


def sample(state_prep=None, shots=0, post_processor=None):
    r"""
    The ``sample`` function allows to take samples from a state that is specified
    by a preparation procedure. This preparation procedure can be supplied via
    a Python function that returns one or more :ref:`QuantumVariables <QuantumVariable>`.

    The samples are returned in the form of a
    `Jax Array <https://jax.readthedocs.io/en/latest/_autosummary/jax.Array.html>`_
    which is shaped according to the ``shots`` parameter. Because of this, shots
    can only be a **static integer** (no dynamic values!). If you want to sample
    with a dynamic shot amount, look into :ref:`expectation_value`.


    Parameters
    ----------
    state_prep : callable
        A function returning one or more :ref:`QuantumVariables <QuantumVariable>`.
        The state from this QuantumVariables will be sampled.
        The state preparation function can only take classical values as arguments.
        This is because a quantum value would need to be copied for each sampling
        iteration, which is prohibited by the no-cloning theorem.
    shots : int
        The amounts of samples to take.
    post_processor : callable, optional
        A function to apply to the samples directly after measuring. By default no post processing is applied.

    Raises
    ------
    Exception
        Tried to sample with dynamic shots value (static integer required)
    Exception
        Tried to sample from state preparation function taking a quantum value

    Returns
    -------
    callable
        A classical, Jax traceable function returning a jax array containing
        the measurement results of each shot.

    Examples
    --------

    We prepare the state

    .. math::

        \ket{\psi} = \frac{1}{\sqrt{2}} \left(\ket{0}\ket{0}\ket{\text{True}} + \ket{k}\ket{k}\ket{\text{True}})\right)

    ::

        from qrisp import *
        from qrisp.jasp import *


        def state_prep(k):
            a = QuantumFloat(4)
            b = QuantumFloat(4)

            qbl = QuantumBool()
            h(qbl)

            with control(qbl[0]):
                a[:] = k

            cx(a, b)

            return a, b

    And subsequently sample from the QuantumFloats:

    ::

        @jaspify
        def main(k):

            sampling_function = sample(state_prep,
                                       shots = 10)

            return sampling_function(k)

        print(main(3))

        # Yields
        # [[3. 3.]
        #  [0. 0.]
        #  [0. 0.]
        #  [3. 3.]
        #  [0. 0.]
        #  [0. 0.]
        #  [3. 3.]
        #  [3. 3.]
        #  [0. 0.]
        #  [0. 0.]]

    To demonstrate the post processing feature, we write a simple post
    processing function:

    ::

        def post_processor(x, y):
            return 2*x + y//2

        @jaspify
        def main(k):

            sampling_function = sample(state_prep,
                                       shots = 10,
                                       post_processor = post_processor)

            return sampling_function(k)

        print(main(4))
        # Yields
        # [10. 10.  0.  0.  0.  0.  0.  0. 10. 10.]

    """

    from qrisp.jasp import qache
    from qrisp.core import QuantumVariable, measure

    if isinstance(state_prep, int):
        shots = state_prep
        state_prep = None

    if state_prep is None:
        return lambda x: sample(x, shots, post_processor=post_processor)

    if post_processor is None:

        def identity(*args):
            if len(args) == 1:
                return args[0]
            return args

        post_processor = identity

    if isinstance(shots, jax.core.Tracer):
        raise Exception(
            "Tried to sample with dynamic shots value (static integer required)"
        )
    elif not isinstance(shots, int):
        raise Exception(
            f"Tried to sample with shots value of non-integer type {type(shots)}"
        )

    # Qache the user function
    @qache
    def user_func(*args):
        return state_prep(*args)

    # This function evaluates the sampling process
    @jax.jit
    def sampling_eval_function(*args, tracerized_shots=0):

        for arg in args:
            if isinstance(arg, QuantumVariable):
                raise Exception(
                    "Tried to sample from state preparation function taking a quantum value"
                )

        # We now construct a loop to collect the samples by
        # inserting the postprocessed measurement result into an array.
        # The following function is the loop body, which is kernelized.
        @quantum_kernel
        def sampling_body_func(i, args):

            acc = args[0]

            # Evaluate the user function
            qv_tuple = user_func(*args[1:])

            if not isinstance(qv_tuple, tuple):
                qv_tuple = (qv_tuple,)

            for qv in qv_tuple:
                if not isinstance(qv, QuantumVariable):
                    raise Exception(
                        "Tried to sample from function not returning a QuantumVariable"
                    )

            # Trace the DynamicQubitArray measurements
            # Since we execute the measurements on the .reg attribute, no decoding
            # is applied. The decoding happens in sampling_helper_2
            @qache
            def sampling_helper_1(*args):
                res_list = []
                for reg in args:
                    res_list.append(measure(reg))
                return tuple(res_list)

            measurement_ints = sampling_helper_1(*[qv.reg for qv in qv_tuple])

            # Trace the decoding
            @jax.jit
            def sampling_helper_2(*meas_ints):
                decoded_values = []
                for j in range(len(qv_tuple)):
                    decoded_values.append(qv_tuple[j].jdecoder(meas_ints[j]))

                if len(qv_tuple) > 1:
                    decoded_values = post_processor(*decoded_values)
                else:
                    decoded_values = post_processor(*decoded_values)

                if isinstance(decoded_values, tuple):
                    # Save the return amount (for more details check the comment of the)
                    # initialization command of return_amount
                    return_amount.append(len(decoded_values))
                    if len(acc.shape) == 1:
                        raise AuxException()

                return decoded_values

            decoded_values = sampling_helper_2(*measurement_ints)
            
            # Insert into the accumulating array
            acc = acc.at[i].set(decoded_values)

            return (acc, *args[1:])

        # This list captures the amount of return values. The strategy here is
        # to initially assume only one QuantumVariable is returned, which is then
        # added to the expectation value accumulator. If more than one is returned,
        # the amount is saved in this list and an exception is raised, which
        # subsequently causes another call but this time with the correct accumulator
        # dimension.

        return_amount = []

        try:
            loop_res = jax.lax.fori_loop(
                0, tracerized_shots, sampling_body_func, (jnp.zeros(shots), *args)
            )
            return loop_res[0]
        except AuxException:
            loop_res = jax.lax.fori_loop(
                0,
                tracerized_shots,
                sampling_body_func,
                (jnp.zeros((shots, return_amount[0])), *args),
            )
            return loop_res[0]

    from qrisp.jasp import terminal_sampling

    def return_function(*args):

        if check_for_tracing_mode():
            return sampling_eval_function(*args, tracerized_shots=shots)
        else:
            return terminal_sampling(state_prep, shots)(*args)

    return return_function


class AuxException(Exception):
    pass



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from functools import lru_cache

from sympy import lambdify, symbols

from jax import make_jaxpr, jit
from jax.lax import fori_loop, cond, while_loop
from jax._src.linear_util import wrap_init
import jax.numpy as jnp

from catalyst.jax_primitives import (
    qinst_p,
    measure_p,
    qextract_p,
    qinsert_p,
    while_p,
    cond_p,
    func_p,
    qdealloc_p,
    qalloc_p,
)

from qrisp.jasp import (
    QuantumPrimitive,
    AbstractQuantumCircuit,
    AbstractQubitArray,
    AbstractQubit,
    eval_jaxpr,
    extract_invalues,
    insert_outvalues,
    quantum_gate_p,
    Measurement_p,
    get_qubit_p,
    get_size_p,
    Jlist,
)

greek_letters = symbols(
    "alpha beta gamma delta epsilon zeta eta theta iota kappa lambda mu nu xi omicron pi rho sigma tau upsilon phi chi psi omega"
)

# Name translator from Qrisp gate naming to Catalyst gate naming
op_name_translation_dic = {
    "cx": "CNOT",
    "cy": "CY",
    "cz": "CZ",
    "crx": "CRX",
    "crz": "CRZ",
    "swap": "SWAP",
    "x": "PauliX",
    "y": "PauliY",
    "z": "PauliZ",
    "h": "Hadamard",
    "rx": "RX",
    "ry": "RY",
    "rz": "RZ",
    "s": "S",
    "t": "T",
    "p": "RZ",
    "u3": "Rot",
}


def catalyst_eqn_evaluator(eqn, context_dic):
    """
    This function serves as the central interface to interpret jasp equations
    into Catalyst primitives.

    Parameters
    ----------
    eqn : jax.core.JaxprEqn
        The equation to be interpreted.
    context_dic : qrisp.jasp.interpreters.ContextDict
        The ContextDict representing the current state of the program by
        translating from variables to values.

    Raises
    ------
    Exception
        Don't know how to interpret QuantumPrimitive.

    Returns
    -------
    bool
        Like other equation processing managers, if this function returns True,
        it implies that the equation should be processed by simply binding the
        primitive.

    """

    # If the equations primitive is a Qrisp primitive, we process it according
    # to one of the implementations below. Otherwise we return True to indicate
    # default interpretation.

    if isinstance(eqn.primitive, QuantumPrimitive):

        invars = eqn.invars
        outvars = eqn.outvars

        if eqn.primitive.name == "jasp.quantum_gate":
            process_op(eqn.params["gate"], invars, outvars, context_dic)
        elif eqn.primitive.name == "jasp.create_qubits":
            process_create_qubits(invars, outvars, context_dic)
        elif eqn.primitive.name == "jasp.get_qubit":
            process_get_qubit(invars, outvars, context_dic)
        elif eqn.primitive.name == "jasp.measure":
            process_measurement(invars, outvars, context_dic)
        elif eqn.primitive.name == "jasp.get_size":
            process_get_size(invars, outvars, context_dic)
        elif eqn.primitive.name == "jasp.slice":
            process_slice(invars, outvars, context_dic)
        elif eqn.primitive.name == "jasp.delete_qubits":
            process_delete_qubits(eqn, context_dic)
        elif eqn.primitive.name == "jasp.reset":
            process_reset(eqn, context_dic)
        elif eqn.primitive.name == "jasp.fuse":
            process_fuse(eqn, context_dic)
        elif eqn.primitive.name == "jasp.create_quantum_kernel":
            qreg = qalloc_p.bind(25)
            insert_outvalues(eqn, context_dic, (qreg, Jlist(jnp.arange(30)[::-1], max_size = 30)))
        elif eqn.primitive.name == "jasp.consume_quantum_kernel":
            invalues = extract_invalues(eqn, context_dic)
            qdealloc_p.bind(invalues[0][0])
        else:
            raise Exception(
                f"Don't know how to process QuantumPrimitive {eqn.primitive}"
            )
    else:
        if eqn.primitive.name == "while":
            return process_while(eqn, context_dic)
        elif eqn.primitive.name == "cond":
            return process_cond(eqn, context_dic)  #
        elif eqn.primitive.name == "pjit":
            process_pjit(eqn, context_dic)
        else:
            return True


def process_create_qubits(invars, outvars, context_dic):

    # The first invar of the create_qubits primitive is an AbstractQuantumCircuit
    # which is represented by an AbstractQreg and an integer
    qreg, free_qubits = context_dic[invars[1]]

    size = context_dic[invars[0]]

    # We create the new QubitArray representation by putting the appropriate tuple
    # in the context_dic

    reg_qubits = Jlist()

    def loop_body(i, val_tuple):
        free_qubits, reg_qubits = val_tuple
        reg_qubits.append(free_qubits.pop())
        return free_qubits, reg_qubits

    @jit
    def make_tracer(x):
        return x

    size = make_tracer(size)

    free_qubits, reg_qubits = fori_loop(0, size, loop_body, (free_qubits, reg_qubits))

    context_dic[outvars[0]] = reg_qubits

    # Furthermore we create the updated AbstractQuantumCircuit representation.
    # The new stack size is the old stask size + the size of the QubitArray
    context_dic[outvars[1]] = (qreg, free_qubits)


def process_delete_qubits(eqn, context_dic):

    # The first invar of the create_qubits primitive is an AbstractQuantumCircuit
    # which is represented by an AbstractQreg and an integer
    qreg, free_qubits = context_dic[eqn.invars[1]]
    reg_qubits = context_dic[eqn.invars[0]]

    # We create the new QubitArray representation by putting the appropriate tuple
    # in the context_dic

    def loop_body(i, val_tuple):
        free_qubits, reg_qubits = val_tuple
        free_qubits.append(reg_qubits.pop())
        return free_qubits, reg_qubits

    @jit
    def make_tracer(x):
        return x

    free_qubits, reg_qubits = fori_loop(
        0, reg_qubits.counter, loop_body, (free_qubits, reg_qubits)
    )

    context_dic[eqn.outvars[0]] = (qreg, free_qubits)


def process_fuse(eqn, context_dic):

    invalues = extract_invalues(eqn, context_dic)

    if isinstance(eqn.invars[0].aval, AbstractQubit) and isinstance(
        eqn.invars[1].aval, AbstractQubit
    ):
        res_qubits = Jlist(invalues)
    elif isinstance(eqn.invars[0].aval, AbstractQubitArray) and isinstance(
        eqn.invars[1].aval, AbstractQubit
    ):
        res_qubits = invalues[0].copy()
        res_qubits.append(invalues[1])
    elif isinstance(eqn.invars[0].aval, AbstractQubit) and isinstance(
        eqn.invars[1].aval, AbstractQubitArray
    ):
        res_qubits = invalues[1].copy()
        res_qubits.prepend(invalues[0])
    else:
        res_qubits = invalues[0].copy()
        res_qubits.extend(invalues[1])

    insert_outvalues(eqn, context_dic, res_qubits)


def process_get_qubit(invars, outvars, context_dic):
    # The get_qubit primitive needs to retrieve the integer that indexes the
    # AbstractQubit in the stack.
    # For that we add the Qubit index (in the QubitArray) to the QubitArray
    # starting index.

    qubit_list = context_dic[invars[0]]
    qubit_index = qubit_list[context_dic[invars[1]]]
    context_dic[outvars[0]] = qubit_index


def process_slice(invars, outvars, context_dic):

    qubit_reg = context_dic[invars[0]]
    start = context_dic[invars[1]]
    stop = context_dic[invars[2]]

    context_dic[outvars[0]] = qubit_reg[start:stop]


def process_get_size(invars, outvars, context_dic):
    # The size is simply the second entry of the QubitArray representation
    context_dic[outvars[0]] = context_dic[invars[0]].counter


def process_op(op, invars, outvars, context_dic):

    # Alias the AbstractQreg
    catalyst_register_tracer = context_dic[invars[-1]][0]

    # We first collect all the Catalyst qubit tracers
    # that are required for the Operation
    qb_vars = []

    # For that we find all the integers (ie. positions) of the participating
    # qubits in the AbstractQreg
    qb_pos = []
    for i in range(op.num_qubits):
        qb_vars.append(invars[i])
        qb_pos.append(context_dic[invars[i]])

    num_qubits = len(qb_pos)

    param_dict = {}

    for i in range(len(op.params)):
        param_dict[op.params[i]] = context_dic[invars[i + op.num_qubits]]

    # Extract the catalyst qubit tracers by using the qextract primitive.
    catalyst_qb_tracers = []
    for i in range(num_qubits):
        catalyst_qb_tracer = qextract_p.bind(catalyst_register_tracer, qb_pos[i])
        catalyst_qb_tracers.append(catalyst_qb_tracer)

    # We can now apply the gate primitive
    res_qbs = exec_qrisp_op(op, catalyst_qb_tracers, param_dict)

    # Finally, we reinsert the qubits and update the register tracer
    for i in range(num_qubits):
        catalyst_register_tracer = qinsert_p.bind(
            catalyst_register_tracer, qb_pos[i], res_qbs[i]
        )

    context_dic[outvars[-1]] = (catalyst_register_tracer, context_dic[invars[-1]][1])


def exec_qrisp_op(op, catalyst_qbs, param_dict):
    # This function takes a Qrisp operation and a list of catalyst qubit tracers
    # and applies the operation to these qubits.

    # If the Operation has a definition, we call this function recursively.
    if op.definition:
        defn = op.definition
        for instr in defn.data:
            qubits = instr.qubits
            qubit_indices = [defn.qubits.index(qb) for qb in qubits]
            temp_catalyst_qbs = [catalyst_qbs[i] for i in qubit_indices]
            res_qbs = exec_qrisp_op(instr.op, temp_catalyst_qbs, param_dict)

            for i in range(len(qubit_indices)):
                catalyst_qbs[qubit_indices[i]] = res_qbs[i]

        return catalyst_qbs

    # Otherwise we simply call the bind method
    else:

        if op.name == "gphase":
            return catalyst_qbs

        if op.name[-3:] == "_dg":
            op_name = op.name[:-3]
            invert = True
        else:
            invert = False
            op_name = op.name

        jax_values = list(param_dict.values())

        param_list = [
            lambdify(greek_letters[: len(op.abstract_params)], expr)(*jax_values)
            for expr in op.params
        ]

        if op_name == "sx":
            op_name = "rx"
            param_list = [jnp.pi / 2]

        if op_name == "u3":

            def qiskit_to_pennylane_rot(theta, phi, lam):
                pl_phi = lam
                pl_theta = theta
                pl_omega = phi
                return pl_phi, pl_theta, pl_omega

            param_list = list(qiskit_to_pennylane_rot(*param_list))

        # param_list = [param_dict[symb] for symb in op.params]

        catalyst_name = op_name_translation_dic[op_name]

        res_qbs = qinst_p.bind(
            *(catalyst_qbs + param_list),
            op=catalyst_name,
            qubits_len=op.num_qubits,
            params_len=len(param_list),
            ctrl_len=0,
            adjoint=invert,
        )

        return res_qbs


def process_measurement(invars, outvars, context_dic):
    # This function treats measurement primitives

    # Retrieve the catalyst register tracer
    catalyst_register_tracer = context_dic[invars[-1]][0]

    # We differentiate between the cases that the measurement is performed on a
    # singular Qubit (returns a bool) or a QubitArray (returns an integer)

    # This case treats the QubitArray situation
    if isinstance(invars[0].aval, AbstractQubitArray):

        # Retrieve the start and the endpoint indices of the QubitArray
        qubit_list = context_dic[invars[0]]

        # The multi measurement logic is outsourced into a dedicated function
        catalyst_register_tracer, meas_res = exec_multi_measurement(
            catalyst_register_tracer, qubit_list
        )

    # The singular Qubit case
    else:

        # Get the position of the Qubit
        qb_pos = context_dic[invars[0]]

        # Get the catalyst Qubit tracer
        catalyst_qb_tracer = qextract_p.bind(catalyst_register_tracer, qb_pos)

        # Call the measurement primitive
        meas_res, res_qb = measure_p.bind(catalyst_qb_tracer)

        # Reinsert the Qubit
        catalyst_register_tracer = qinsert_p.bind(
            catalyst_register_tracer, qb_pos, res_qb
        )

    # Insert the result values into the context dict
    context_dic[outvars[1]] = (catalyst_register_tracer, context_dic[invars[1]][1])
    context_dic[outvars[0]] = meas_res


def exec_multi_measurement(catalyst_register, qubit_list):
    # This function performs the measurement of multiple qubits at once, returning
    # an integer. The qubits to be measured sit in one consecutive interval,
    # starting at index "start" and ending at "stop".

    # We realize this feature using the Catalyst while loop primitive.

    # For that we need to define the loop body and the condition for breaking the
    # loop

    def loop_body(i, acc, reg, static_qubit_array, list_size):

        qb_index = static_qubit_array[i]
        qb = qextract_p.bind(reg, qb_index)
        res_bl, res_qb = measure_p.bind(qb)
        reg = qinsert_p.bind(reg, qb_index, res_qb)
        acc = acc + (jnp.asarray(1, dtype="int64") << i) * res_bl
        i += jnp.asarray(1, dtype="int64")
        return i, acc, reg, static_qubit_array, list_size

    def cond_body(i, acc, reg, static_qubit_array, list_size):
        return i < list_size

    static_qubit_array, list_size = qubit_list.array, qubit_list.counter

    # Turn them into jaxpr
    zero = jnp.asarray(0, dtype="int64")
    loop_jaxpr = make_jaxpr(loop_body)(
        zero, zero, catalyst_register.aval, static_qubit_array, list_size
    )
    cond_jaxpr = make_jaxpr(cond_body)(
        zero, zero, catalyst_register.aval, static_qubit_array, list_size
    )

    # Bind the while loop primitive
    while_res = while_p.bind(
        *(zero, zero, catalyst_register, static_qubit_array, list_size),
        cond_jaxpr=cond_jaxpr,
        body_jaxpr=loop_jaxpr,
        cond_nconsts=0,
        body_nconsts=0,
        nimplicit=0,
        preserve_dimensions=True,
    )

    i, acc, reg, static_qubit_array, list_size = while_res

    return reg, acc


def process_while(eqn, context_dic):

    body_jaxpr = ensure_conversion(eqn.params["body_jaxpr"])
    cond_jaxpr = ensure_conversion(eqn.params["cond_jaxpr"])

    invalues = extract_invalues(eqn, context_dic)

    flattened_invalues = flatten_signature(invalues, eqn.invars)

    k = eqn.params["body_nconsts"]
    new_body_nconsts = len(flatten_signature(invalues[:k], eqn.invars[:k]))

    k = eqn.params["cond_nconsts"]
    new_cond_nconsts = len(flatten_signature(invalues[:k], eqn.invars[:k]))
    
    outvalues = while_p.bind(
        *flattened_invalues,
        cond_jaxpr=cond_jaxpr,
        body_jaxpr=body_jaxpr,
        cond_nconsts=new_cond_nconsts,
        body_nconsts=new_body_nconsts,
        nimplicit=0,
        preserve_dimensions=True,
    )

    outvalues = list(outvalues)
    unflattened_outvalues = unflatten_signature(outvalues, eqn.outvars)

    insert_outvalues(eqn, context_dic, unflattened_outvalues)


def process_cond(eqn, context_dic):

    branch_list = []

    for i in range(len(eqn.params["branches"])):
        branch_list.append(ensure_conversion(eqn.params["branches"][i]))

    invalues = extract_invalues(eqn, context_dic)

    if isinstance(eqn.invars[-1].aval, AbstractQuantumCircuit):

        if len(branch_list) > 2:
            raise Exception(
                "Converting cond primitive with more than 2 branches to Catalyst is currently not supported."
            )

        # Contrary to the jax cond primitive, the catalyst cond primitive
        # wants a bool
        invalues[0] = jnp.asarray(invalues[0], bool)

        flattened_invalues = flatten_signature(invalues, eqn.invars)

        outvalues = cond_p.bind(
            *flattened_invalues, branch_jaxprs=branch_list[::-1], nimplicit_outputs=0
        )

        unflattened_outvalues = unflatten_signature(outvalues, eqn.outvars)
    else:

        unflattened_outvalues = eqn.primitive.bind(
            *invalues, branches=branch_list, linear=False
        )

    insert_outvalues(eqn, context_dic, unflattened_outvalues)


@lru_cache(maxsize=int(1e5))
def get_traced_fun(jaxpr):

    catalyst_jaxpr = ensure_conversion(jaxpr)

    from jax.core import eval_jaxpr

    @jit
    def jitted_fun(*args):
        return eval_jaxpr(catalyst_jaxpr.jaxpr, [], *args)

    return jitted_fun


def process_pjit(eqn, context_dic):

    invalues = extract_invalues(eqn, context_dic)

    flattened_invalues = flatten_signature(invalues, eqn.invars)

    jaxpr = eqn.params["jaxpr"]
    traced_fun = get_traced_fun(jaxpr)

    outvalues = func_p.bind(wrap_init(traced_fun, debug_info = eqn.params["jaxpr"].jaxpr.debug_info),
                            *flattened_invalues, 
                            fn=traced_fun,)

    outvalues = list(outvalues)

    unflattened_outvalues = unflatten_signature(outvalues, eqn.outvars)

    insert_outvalues(eqn, context_dic, unflattened_outvalues)


# Function to reset and delete a qubit array
def reset_qubit_array(qb_array, abs_qc):
    from qrisp.circuit import XGate

    def body_func(arg_tuple):

        qb_array, i, abs_qc = arg_tuple

        abs_qb = get_qubit_p.bind(qb_array, i)
        meas_bl, abs_qc = Measurement_p.bind(abs_qb, abs_qc)

        def true_fun(arg_tuple):
            qb, abs_qc = arg_tuple
            abs_qc = quantum_gate_p.bind(qb, abs_qc, gate = XGate())
            return (qb, abs_qc)

        def false_fun(arg_tuple):
            return arg_tuple

        qb, abs_qc = cond(meas_bl, true_fun, false_fun, (abs_qb, abs_qc))

        i += 1

        return (qb_array, i, abs_qc)

    def cond_fun(arg_tuple):
        return arg_tuple[-2] < get_size_p.bind(arg_tuple[0])

    qb_array, i, abs_qc = while_loop(
        cond_fun, body_func, (qb_array, jnp.array(0, dtype=jnp.int64), abs_qc)
    )

    return abs_qc


reset_jaxpr = make_jaxpr(reset_qubit_array)(
    AbstractQubitArray(), AbstractQuantumCircuit()
)


def process_reset(eqn, context_dic):

    invalues = extract_invalues(eqn, context_dic)
    outvalues = eval_jaxpr(reset_jaxpr.jaxpr, eqn_evaluator=catalyst_eqn_evaluator)(
        *invalues
    )
    insert_outvalues(eqn, context_dic, outvalues)


def ensure_conversion(jaxpr):
    from qrisp.jasp.evaluation_tools.catalyst_interface import jaspr_to_catalyst_jaxpr

    return jaspr_to_catalyst_jaxpr(jaxpr)


def flatten_signature(values, variables):
    values = list(values)
    flattened_values = []
    for i in range(len(variables)):
        var = variables[i]
        value = values.pop(0)
        if isinstance(var.aval, AbstractQuantumCircuit):
            flattened_values.extend((value[0], *value[1].flatten()[0]))
        elif isinstance(var.aval, AbstractQubitArray):
            flattened_values.extend(value.flatten()[0])
        else:
            flattened_values.append(value)

    return flattened_values


def unflatten_signature(values, variables):
    values = list(values)
    unflattened_values = []
    for var in variables:
        if isinstance(var.aval, AbstractQuantumCircuit):
            catalyst_register_tracer = values.pop(0)
            jlist_tuple = (values.pop(0), values.pop(0))
            unflattened_values.append(
                (catalyst_register_tracer, Jlist.unflatten([], jlist_tuple))
            )
        elif isinstance(var.aval, AbstractQubitArray):
            jlist_tuple = (values.pop(0), values.pop(0))
            unflattened_values.append(Jlist.unflatten([], jlist_tuple))
        else:
            unflattened_values.append(values.pop(0))

    return unflattened_values



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from functools import lru_cache

import numpy as np

from jax import make_jaxpr, jit, debug
from jax.extend.core import ClosedJaxpr, Literal
from jax.lax import fori_loop, cond, while_loop, switch
import jax.numpy as jnp

from qrisp.circuit import PTControlledOperation
from qrisp.jasp import (
    QuantumPrimitive,
    AbstractQuantumCircuit,
    AbstractQubitArray,
    AbstractQubit,
    eval_jaxpr,
    Jaspr,
    extract_invalues,
    insert_outvalues,
    Jlist,
)


def cl_func_eqn_evaluator(eqn, context_dic):

    # If the equations primitive is a Qrisp primitive, we process it according
    # to one of the implementations below. Otherwise we return True to indicate
    # default interpretation.
    if isinstance(eqn.primitive, QuantumPrimitive):

        invars = eqn.invars
        outvars = eqn.outvars

        if eqn.primitive.name == "jasp.quantum_gate":
            process_op(eqn.params["gate"], invars, outvars, context_dic)
        elif eqn.primitive.name == "jasp.create_qubits":
            process_create_qubits(invars, outvars, context_dic)
        elif eqn.primitive.name == "jasp.get_qubit":
            process_get_qubit(invars, outvars, context_dic)
        elif eqn.primitive.name == "jasp.measure":
            process_measurement(invars, outvars, context_dic)
        elif eqn.primitive.name == "jasp.get_size":
            process_get_size(invars, outvars, context_dic)
        elif eqn.primitive.name == "jasp.slice":
            process_slice(invars, outvars, context_dic)
        elif eqn.primitive.name == "jasp.delete_qubits":
            process_delete_qubits(eqn, context_dic)
        elif eqn.primitive.name == "jasp.reset":
            process_reset(eqn, context_dic)
        elif eqn.primitive.name == "jasp.fuse":
            process_fuse(eqn, context_dic)
        else:
            raise Exception(
                f"Don't know how to process QuantumPrimitive {eqn.primitive}"
            )
    else:
        if eqn.primitive.name == "while":
            return process_while(eqn, context_dic)
        elif eqn.primitive.name == "cond":
            return process_cond(eqn, context_dic)  #
        elif eqn.primitive.name == "pjit":
            process_pjit(eqn, context_dic)
        else:
            return True


def process_create_qubits(invars, outvars, context_dic):

    # The first invar of the create_qubits primitive is an AbstractQuantumCircuit
    # which is represented by an AbstractQreg and an integer
    qreg, free_qubits = context_dic[invars[1]]

    size = context_dic[invars[0]]

    # We create the new QubitArray representation by putting the appropriate tuple
    # in the context_dic

    reg_qubits = Jlist()

    def loop_body(i, val_tuple):
        reg_qubits, free_qubits = val_tuple
        reg_qubits.append(free_qubits.pop())
        return reg_qubits, free_qubits

    @jit
    def make_tracer(x):
        return x

    size = make_tracer(size)

    reg_qubits, free_qubits = fori_loop(0, size, loop_body, (reg_qubits, free_qubits))

    context_dic[outvars[0]] = reg_qubits

    # Furthermore we create the updated AbstractQuantumCircuit representation.
    # The new stack size is the old stask size + the size of the QubitArray
    context_dic[outvars[1]] = (qreg, free_qubits)


def process_fuse(eqn, context_dic):

    invalues = extract_invalues(eqn, context_dic)

    if isinstance(eqn.invars[0].aval, AbstractQubit) and isinstance(
        eqn.invars[1].aval, AbstractQubit
    ):
        res_qubits = Jlist(invalues)
    elif isinstance(eqn.invars[0].aval, AbstractQubitArray) and isinstance(
        eqn.invars[1].aval, AbstractQubit
    ):
        res_qubits = invalues[0].copy()
        res_qubits.append(invalues[1])
    elif isinstance(eqn.invars[0].aval, AbstractQubit) and isinstance(
        eqn.invars[1].aval, AbstractQubitArray
    ):
        res_qubits = invalues[1].copy()
        res_qubits.prepend(invalues[0])
    else:
        res_qubits = invalues[0].copy()
        res_qubits.extend(invalues[1])

    insert_outvalues(eqn, context_dic, res_qubits)


def process_get_qubit(invars, outvars, context_dic):
    # The get_qubit primitive needs to retrieve the integer that indexes the
    # AbstractQubit in the stack.
    # For that we add the Qubit index (in the QubitArray) to the QubitArray
    # starting index.

    reg_qubits = context_dic[invars[0]]
    index = context_dic[invars[1]]
    context_dic[outvars[0]] = reg_qubits[index]


def process_slice(invars, outvars, context_dic):

    reg_qubits = context_dic[invars[0]]
    start = context_dic[invars[1]]
    stop = context_dic[invars[2]]

    context_dic[outvars[0]] = reg_qubits[start:stop]


def process_get_size(invars, outvars, context_dic):
    # The size is simply the second entry of the QubitArray representation
    context_dic[outvars[0]] = context_dic[invars[0]].counter


def process_op(op, invars, outvars, context_dic):

    # B
    bit_array = context_dic[invars[-1]][0]

    # For that we find all the integers (ie. positions) of the participating
    # qubits in the AbstractQreg
    qb_pos = []
    for i in range(op.num_qubits):
        qb_pos.append(context_dic[invars[i]])

    if op.name in ["x", "swap"]:
        ctrl_state = ""
    elif isinstance(op, PTControlledOperation) and op.base_operation.name in [
        "x",
        "y",
        "z",
        "pt2cx",
        "swap"
    ]:
        if op.base_operation.name in ["z", "rz", "t", "s", "t_dg", "s_dg", "p"]:
            context_dic[outvars[-1]] = context_dic[invars[-1]]
            return
        ctrl_state = op.ctrl_state
    elif op.name in ["z", "rz", "t", "s", "t_dg", "s_dg", "p"]:
        context_dic[outvars[-1]] = context_dic[invars[-1]]
        return
    else:
        raise Exception(f"Classical function simulator can't process gate {op.name}")

    if "swap" in op.name:
        swapped_qb_pos = list(qb_pos)
        swapped_qb_pos[-1], swapped_qb_pos[-2] = swapped_qb_pos[-2], swapped_qb_pos[-1]
        ctrl_state = ctrl_state + "1"
        
        bit_array = cl_multi_cx(bit_array, ctrl_state, swapped_qb_pos)
        bit_array = cl_multi_cx(bit_array, ctrl_state, qb_pos)
        bit_array = cl_multi_cx(bit_array, ctrl_state, swapped_qb_pos)
    else:
        bit_array = cl_multi_cx(bit_array, ctrl_state, qb_pos)

    context_dic[outvars[-1]] = (bit_array, context_dic[invars[-1]][1])


def cl_multi_cx(bit_array, ctrl_state, bit_pos):

    flip = jnp.asarray(1, dtype=bit_array.dtype)
    for i in range(len(bit_pos) - 1):
        if ctrl_state[i] == "1":
            flip = flip & (get_bit_array(bit_array, bit_pos[i]))
        else:
            flip = flip & (~get_bit_array(bit_array, bit_pos[i]))

    bit_array = conditional_bit_flip_bit_array(bit_array, bit_pos[-1], flip)
    # bit_array = bit_array.at[bit_pos[-1]].set((bit_array[bit_pos[-1]] ^ flip), mode = "promise_in_bounds")

    return bit_array


def process_measurement(invars, outvars, context_dic):

    # This function treats measurement primitives

    # Retrieve the catalyst register tracer
    bit_array = context_dic[invars[-1]][0]

    # We differentiate between the cases that the measurement is performed on a
    # singular Qubit (returns a bool) or a QubitArray (returns an integer)

    # This case treats the QubitArray situation
    if isinstance(invars[0].aval, AbstractQubitArray):

        # Retrieve the start and the endpoint indices of the QubitArray
        qubit_reg = context_dic[invars[0]]

        # The multi measurement logic is outsourced into a dedicated function
        bit_array, meas_res = exec_multi_measurement(bit_array, qubit_reg)

    # The singular Qubit case
    else:
        # Get the boolean value of the Qubit
        meas_res = get_bit_array(bit_array, context_dic[invars[0]])
        # meas_res = bit_array[context_dic[invars[1]]]

    # Insert the result values into the context dict
    context_dic[outvars[1]] = (bit_array, context_dic[invars[1]][1])
    context_dic[outvars[0]] = meas_res


def exec_multi_measurement(bit_array, qubit_reg):
    # This function performs the measurement of multiple qubits at once, returning
    # an integer. The qubits to be measured sit in one consecutive interval,
    # starting at index "start" and ending at "stop".

    # We realize this feature using the Catalyst while loop primitive.

    # For that we need to define the loop body and the condition for breaking the
    # loop

    def loop_body(i, arg_tuple):
        acc, bit_array, qubit_reg = arg_tuple
        qb_index = qubit_reg[i]
        res_bl = get_bit_array(bit_array, qb_index)
        acc = acc + (jnp.asarray(res_bl, dtype="int64") << i)
        return (acc, bit_array, qubit_reg)

    acc, bit_array, qubit_reg = fori_loop(
        0, qubit_reg.counter, loop_body, (0, bit_array, qubit_reg)
    )

    return bit_array, acc


def process_while(eqn, context_dic):
    for invar in eqn.invars:
        if isinstance(invar.aval, AbstractQuantumCircuit):
            break
    else:
        return True

    body_jaxpr = eqn.params["body_jaxpr"]
    cond_jaxpr = eqn.params["cond_jaxpr"]
    overall_constant_amount= eqn.params["body_nconsts"] + eqn.params["cond_nconsts"]

    invalues = extract_invalues(eqn, context_dic)

    if isinstance(body_jaxpr.jaxpr.invars[-1].aval, AbstractQuantumCircuit):
        converted_body_jaxpr = jaspr_to_cl_func_jaxpr(
            body_jaxpr.jaxpr, invalues[-1][0].shape[0] * 64
        )
    if isinstance(cond_jaxpr.jaxpr.invars[-1].aval, AbstractQuantumCircuit):
        converted_cond_jaxpr = jaspr_to_cl_func_jaxpr(
            cond_jaxpr.jaxpr, invalues[-1][0].shape[0] * 64
        )

    def body_fun(args):
        
        constants = args[eqn.params["cond_nconsts"]:overall_constant_amount]
        carries = args[overall_constant_amount:]
        
        flattened_invalues = flatten_signature(constants + carries, body_jaxpr.jaxpr.invars)
        
        body_res = eval_jaxpr(converted_body_jaxpr)(*(flattened_invalues))
        
        unflattened_body_outvalues = unflatten_signature(body_res, body_jaxpr.jaxpr.outvars)
        
        return list(args[:overall_constant_amount]) + list(unflattened_body_outvalues)

    def cond_fun(args):
        
        constants = args[:eqn.params["cond_nconsts"]]
        carries = args[overall_constant_amount:]
        
        flattened_invalues = flatten_signature(constants + carries, cond_jaxpr.jaxpr.invars)
        
        return eval_jaxpr(converted_cond_jaxpr)(*flattened_invalues)

    outvalues = while_loop(cond_fun, body_fun, invalues)[eqn.params["body_nconsts"]:]
    
    insert_outvalues(eqn, context_dic, outvalues)


def process_cond(eqn, context_dic):

    invalues = extract_invalues(eqn, context_dic)

    branch_list = []

    for i in range(len(eqn.params["branches"])):
        
        converted_jaxpr = ensure_conversion(
            eqn.params["branches"][i].jaxpr, invalues[1:]
        )
        branch_list.append(eval_jaxpr(converted_jaxpr))

    flattened_invalues = flatten_signature(invalues, eqn.invars)
    outvalues = switch(flattened_invalues[0], branch_list, *flattened_invalues[1:])

    if isinstance(outvalues, tuple):
        unflattened_outvalues = unflatten_signature(outvalues, eqn.outvars)
    else:
        unflattened_outvalues = (outvalues,)

    insert_outvalues(eqn, context_dic, unflattened_outvalues)


@lru_cache(maxsize=int(1e5))
def get_traced_fun(jaxpr, bit_array_size):

    from jax.core import eval_jaxpr

    if isinstance(jaxpr, Jaspr):
        cl_func_jaxpr = jaspr_to_cl_func_jaxpr(jaxpr, bit_array_size)
    else:
        cl_func_jaxpr = jaxpr

    @jit
    def jitted_fun(*args):
        return eval_jaxpr(cl_func_jaxpr.jaxpr, cl_func_jaxpr.consts, *args)

    return jitted_fun


def process_pjit(eqn, context_dic):

    invalues = extract_invalues(eqn, context_dic)
    
    if eqn.params["name"] in ["gidney_mcx", "gidney_mcx_inv"]:
        unflattened_outvalues = [
            (cl_multi_cx(invalues[-1][0], "11", invalues[:-1]), invalues[-1][1])
        ]
    elif eqn.params["name"] in ["gidney_CCCX", "gidney_CCCX_dg"]:
        controls = [invalues[0][i] for i in range(3)]
        unflattened_outvalues = [
            (cl_multi_cx(invalues[-1][0], "111",  controls + [invalues[1]]), invalues[-1][1])
        ]
    else:
        flattened_invalues = []
        for value in invalues:
            if isinstance(value, tuple):
                flattened_invalues.extend(value)
            else:
                flattened_invalues.append(value)

        jaxpr = eqn.params["jaxpr"]

        if isinstance(jaxpr, Jaspr):
            bit_array_padding = invalues[-1][0].shape[0] * 64
        else:
            bit_array_padding = 0

        traced_fun = get_traced_fun(jaxpr, bit_array_padding)

        flattened_invalues = flatten_signature(invalues, eqn.invars)
        outvalues = traced_fun(*flattened_invalues)
        unflattened_outvalues = unflatten_signature(outvalues, eqn.outvars)

    insert_outvalues(eqn, context_dic, unflattened_outvalues)


def process_delete_qubits(eqn, context_dic):
    invalues = extract_invalues(eqn, context_dic)

    bit_array = invalues[-1][0]
    free_qubits = invalues[-1][1]
    qubit_reg = invalues[0]

    def true_fun():
        debug.print("WARNING: Faulty uncomputation found during simulation.")

    def false_fun():
        return

    def loop_body(i, value_tuple):
        qubit_reg, bit_array, free_qubits = value_tuple

        qubit = qubit_reg.pop()
        cond(get_bit_array(bit_array, qubit), true_fun, false_fun)
        free_qubits.append(qubit)

        return qubit_reg, bit_array, free_qubits

    qubit_reg, bit_array, free_qubits = fori_loop(
        0, qubit_reg.counter, loop_body, (qubit_reg, bit_array, free_qubits)
    )

    outvalues = (bit_array, free_qubits)

    insert_outvalues(eqn, context_dic, outvalues)


def process_reset(eqn, context_dic):

    invalues = extract_invalues(eqn, context_dic)

    bit_array = invalues[-1][0]

    if isinstance(eqn.invars[0].aval, AbstractQubitArray):

        start = invalues[0][0]
        stop = start + invalues[0][1]

        def loop_body(i, bit_array):
            bit_array = conditional_bit_flip_bit_array(
                bit_array, i, get_bit_array(bit_array, i)
            )
            return bit_array

        bit_array = fori_loop(start, stop, loop_body, bit_array)

    else:

        bit_array = conditional_bit_flip_bit_array(
            bit_array, invalues[0], get_bit_array(bit_array, invalues[0])
        )

    outvalues = (bit_array, invalues[-1][1])
    insert_outvalues(eqn, context_dic, outvalues)


def jaspr_to_cl_func_jaxpr(jaspr, bit_array_padding):

    args = []
    for invar in jaspr.invars:
        if isinstance(invar.aval, AbstractQuantumCircuit):
            args.append(
                (
                    jnp.zeros(int(np.ceil(bit_array_padding / 64)), dtype=jnp.uint64),
                    Jlist(jnp.arange(bit_array_padding), max_size=bit_array_padding),
                )
            )
        elif isinstance(invar.aval, AbstractQubitArray):
            args.append(Jlist())
        elif isinstance(invar.aval, AbstractQubit):
            args.append(jnp.asarray(0, dtype="int64"))
        elif isinstance(invar, Literal):
            if isinstance(invar.val, int):
                args.append(jnp.asarray(invar.val, dtype="int64"))
            if isinstance(invar.val, float):
                args.append(jnp.asarray(invar.val, dtype="float64"))
        else:
            args.append(invar.aval)

    return make_jaxpr(eval_jaxpr(jaspr, eqn_evaluator=cl_func_eqn_evaluator))(
        *args
    )


def conditional_bit_flip_bit_array(bit_array, index, condition):
    index = jnp.uint64(index)
    array_index = index >> 6
    int_index = index ^ (array_index << 6)
    set_value = bit_array[array_index] ^ (condition << int_index)
    # set_value = bit_array[index]
    return bit_array.at[array_index].set(set_value)


conditional_bit_flip_bit_array = jit(conditional_bit_flip_bit_array, donate_argnums=0)


def get_bit_array(bit_array, index):
    index = jnp.uint64(index)
    array_index = index >> 6
    int_index = index ^ (array_index << 6)
    return (bit_array[array_index] >> int_index) & 1


def unflatten_signature(values, variables):
    values = list(values)
    unflattened_values = []
    for var in variables:
        if isinstance(var.aval, AbstractQuantumCircuit):
            bit_array = values.pop(0)
            jlist_tuple = (values.pop(0), values.pop(0))
            unflattened_values.append((bit_array, Jlist.unflatten([], jlist_tuple)))
        elif isinstance(var.aval, AbstractQubitArray):
            jlist_tuple = (values.pop(0), values.pop(0))
            unflattened_values.append(Jlist.unflatten([], jlist_tuple))
        else:
            unflattened_values.append(values.pop(0))

    return unflattened_values


def flatten_signature(values, variables):
    values = list(values)
    flattened_values = []
    for i in range(len(variables)):
        var = variables[i]
        value = values.pop(0)
        if isinstance(var.aval, AbstractQuantumCircuit):
            flattened_values.extend((value[0], *value[1].flatten()[0]))
        elif isinstance(var.aval, AbstractQubitArray):
            flattened_values.extend(value.flatten()[0])
        else:
            flattened_values.append(value)

    return flattened_values


def ensure_conversion(jaxpr, invalues):

    bit_array_padding = 0
    convert = False
    for i in range(len(jaxpr.invars)):
        invar = jaxpr.invars[i]
        if isinstance(
            invar.aval, (AbstractQuantumCircuit, AbstractQubitArray, AbstractQubit)
        ):
            convert = True
            if isinstance(invar.aval, AbstractQuantumCircuit):
                bit_array_padding = invalues[i][0].shape[0] * 64

    if convert:
        return jaspr_to_cl_func_jaxpr(jaxpr, bit_array_padding)
    return ClosedJaxpr(jaxpr, [])



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import types
from dataclasses import dataclass
from typing import Callable, Optional

import numpy as np

from qrisp import QuantumSession
from qrisp.circuit import ControlledOperation, Operation, QuantumCircuit

has_pennylane = True

try:
    import pennylane as qml
except (ModuleNotFoundError, ImportError) as import_error:
    has_pennylane = False

"""
TODO:

-- Implement all gates listed in `scr/qrisp/circuit/standard_operations.py`
-- Add support for control-flow operations (if, while, for)
-- Implement complex parameters handling for gates that support them.

"""


@dataclass
class QMLGateDescriptor:
    """Descriptor for mapping Qrisp operations to PennyLane gates."""

    gate_class: Callable
    param_fn: Callable = lambda op: op.params


if has_pennylane:
    QRISP_PL_BASE_MAP = {
        "x": QMLGateDescriptor(qml.X),
        "y": QMLGateDescriptor(qml.Y),
        "z": QMLGateDescriptor(qml.Z),
        "h": QMLGateDescriptor(qml.Hadamard),
        "rx": QMLGateDescriptor(qml.RX),
        "ry": QMLGateDescriptor(qml.RY),
        "rz": QMLGateDescriptor(qml.RZ),
        "p": QMLGateDescriptor(qml.PhaseShift),
        "u1": QMLGateDescriptor(qml.RZ),
        "u3": QMLGateDescriptor(qml.U3),
        "rxx": QMLGateDescriptor(qml.IsingXX),
        "ryy": QMLGateDescriptor(qml.IsingYY),
        "rzz": QMLGateDescriptor(qml.IsingZZ),
        # qml.IsingXY(phi, ...) is equivalent to XXYYGate(phi, beta=np.pi)
        "swap": QMLGateDescriptor(qml.SWAP),
        "s": QMLGateDescriptor(qml.S),
        "t": QMLGateDescriptor(qml.T),
        "id": QMLGateDescriptor(qml.Identity, param_fn=lambda _: []),
        "sx": QMLGateDescriptor(
            qml.RX, param_fn=lambda _: [np.pi / 2]
        ),  # SXGate() -> qml.RX(pi/2)
        "gphase": QMLGateDescriptor(
            qml.GlobalPhase, param_fn=lambda op: [-op.params[0]]
        ),  # GPhaseGate(phi) -> qml.GlobalPhase(-phi)
        "r": QMLGateDescriptor(
            qml.Rot, param_fn=lambda op: [-op.params[1], -op.params[0], op.params[1]]
        ),  # RGate(theta, phi) -> qml.Rot(-phi, -theta, phi)
        "measure": QMLGateDescriptor(qml.measurements.MidMeasureMP),
    }


def _extract_name(name: str) -> tuple[str, bool]:
    """Extract the operation name and check if it's inverted."""
    is_inverse = name.endswith("_dg")
    name = name[:-3] if is_inverse else name
    return name, is_inverse


def _create_qml_instruction(op: Operation) -> tuple[QMLGateDescriptor, bool, bool]:
    """
    Create a PennyLane instruction from a Qrisp operation.

    Parameters
    ----------
    op : Operation
        The Qrisp operation to convert.

    Returns
    -------
    tuple[QMLGateDescriptor, bool, bool]
        A tuple containing the QMLGateDescriptor corresponding to the Qrisp operation,
        a boolean indicating if the operation is inverted, and a boolean indicating
        if the operation is controlled. For controlled Qrisp operations, it
        returns the base QMLGateDescriptor.

    """

    name, is_inverse = _extract_name(op.name)

    if name in QRISP_PL_BASE_MAP:
        return QRISP_PL_BASE_MAP[name], is_inverse, False

    if isinstance(op, ControlledOperation):
        base_name, is_inverse = _extract_name(op.base_operation.name)
        if base_name in QRISP_PL_BASE_MAP:
            return QRISP_PL_BASE_MAP[base_name], is_inverse, True

    raise NotImplementedError(
        f"Operation '{op.name}' is not supported in the PennyLane converter."
    )


def _is_nested_circuit(op: Operation) -> bool:
    """Return True if this Qrisp operation is a subcircuit."""
    return (
        op.definition is not None
        and not isinstance(op, ControlledOperation)
        and op.name not in QRISP_PL_BASE_MAP
    )


def _evaluate_abstract_params(params: list, subs_dic: dict) -> list:
    """Return parameters with substitution for abstract parameters."""
    out = []
    for p in params:
        if hasattr(p, "subs"):
            p = p.subs(subs_dic)
        out.append(np.float64(p))
    return out


def _process_qrisp_circuit(
    qc: QuantumCircuit | QuantumSession, wire_map: dict, subs_dic: dict | None
) -> None:
    """Recursively process a Qrisp circuit into PennyLane operations."""

    for data in qc.data:
        op = data.op

        if op.name in {"qb_alloc", "qb_dealloc"}:
            continue

        if _is_nested_circuit(op):
            sub_qc = op.definition
            sub_qubits = data.qubits
            sub_wire_map = {
                inner.identifier: wire_map[outer.identifier]
                for inner, outer in zip(sub_qc.qubits, sub_qubits)
            }
            _process_qrisp_circuit(sub_qc, sub_wire_map, subs_dic)
            continue

        qml_gate_desc, is_inverse, is_controlled = _create_qml_instruction(op)
        qml_wires = [wire_map[qb.identifier] for qb in data.qubits]

        n_ctrls = len(op.ctrl_state) if is_controlled else 0
        controls, targets = qml_wires[:n_ctrls], qml_wires[n_ctrls:]

        qml_params = qml_gate_desc.param_fn(op)

        if subs_dic is not None:
            qml_params = _evaluate_abstract_params(qml_params, subs_dic)

        if qml_gate_desc.gate_class is qml.measurements.MidMeasureMP:
            qml.measure(wires=targets)
            continue

        with qml.QueuingManager.stop_recording():

            qml_op = qml_gate_desc.gate_class(*qml_params, wires=targets)

            if is_inverse:
                qml_op = qml.adjoint(qml_op)

            if is_controlled:
                qml_op = qml.ctrl(
                    op=qml_op,
                    control=controls,
                    control_values=[int(bit) for bit in op.ctrl_state],
                )

        qml.apply(qml_op)


def qml_converter(qc: QuantumCircuit | QuantumSession) -> types.FunctionType:
    """
    Convert a Qrisp quantum circuit to a PennyLane quantum function.

    Parameters
    ----------
    qc : QuantumCircuit | QuantumSession
        The Qrisp quantum circuit to be converted.

    Returns
    -------
    types.FunctionType
        A PennyLane quantum function reproducing the Qrisp circuit.

    """

    if not has_pennylane:
        raise ImportError(
            "This feature requires pennylane, a library for quantum computing and "
            "quantum machine learning. It can be installed with:\n\npip install pennylane"
        )  # pragma: no cover

    def circuit(
        wires: Optional[qml.wires.WiresLike] = None, subs_dic: Optional[dict] = None
    ) -> None:
        """
        PennyLane quantum function representing the Qrisp circuit.

        This function should be used within a PennyLane QNode context.

        Parameters
        ----------
        wires : qml.wires.WiresLike, optional
            The wires to be used in the PennyLane circuit. If None, the qubit identifiers
            from the Qrisp circuit are used.

        subs_dic : dict, optional
            A dictionary for substituting abstract parameters with concrete values.

        """

        if wires is None:
            wire_map = {qb.identifier: qb.identifier for qb in qc.qubits}
        else:
            wire_map = {qb.identifier: wires[i] for i, qb in enumerate(qc.qubits)}

        _process_qrisp_circuit(qc, wire_map, subs_dic)

    return circuit



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

# This file implements the carry generation art of the algorithm presented here:
# https://arxiv.org/abs/2304.02921

# For this we implement the Brent-Kung-Tree in a recursive manner. See:
# https://web.stanford.edu/class/archive/ee/ee371/ee371.1066/lectures/lect_04.pdf Page 13

from qrisp.core.gate_application_functions import x, cx
from qrisp.misc.utility import check_if_fresh
from qrisp.alg_primitives.arithmetic.adders.gidney import gidney_adder, cq_gidney_adder
from qrisp.alg_primitives.arithmetic.adders.incrementation import lin_incr

# Both of these functions are the semi-classical versions of the ones found in
# qrisp.alg_primitives.arithmetic.adders.qcla.quantum_quantum.qq_sum_path


def cq_sum_path(a, b, c, R, ctrl=None):

    if len(a) < len(b):
        a = a + "0" * (len(b) - len(a))

    if R == 1:
        # If R = 1, we are in the case of Drapers QCLA:
        # https://arxiv.org/abs/quant-ph/0406142
        # We can use the formula
        # S = A (+) B (+) C
        cx(c[:-1], b[1:])

        for i in range(len(a)):
            if a[i] == "1":
                x(b)
    else:
        # Execute addition using the corresponding carry values
        for i in range(len(c) + 1):

            # Determine the radix qubits to perform the addition on
            a_block = a[R * i : R * (i + 1)]
            b_block = b[R * i : R * (i + 1)]

            if i == 0 and a_block != len(a_block) * "0":
                gidney_adder(a_block, b_block, ctrl=ctrl)
            elif a_block != len(a_block) * "0":
                gidney_adder(a_block, b_block, c_in=c[i - 1], ctrl=ctrl)
            elif not check_if_fresh([c[i - 1]], c[i - 1].qs()):
                lin_incr(b_block, c_in=c[i - 1])


def cq_sum_path_direct_uncomputation(a, b, c, R, ctrl=None):

    if len(a) < len(b):
        a = a + "0" * (len(b) - len(a))

    if R == 1 and False:
        # If R = 1, we are in the case of Drapers QCLA:
        # https://arxiv.org/abs/quant-ph/0406142
        # We can use the formula
        # S = A (+) B (+) C
        for i in range(len(c)):
            cx(c[i], b[i + 1])

        for i in range(len(a)):
            if a[i] == "1":
                if ctrl is None:
                    x(b[i])
                else:
                    cx(ctrl, b[i])
    else:
        # Execute addition using the corresponding carry values
        for i in range(len(c) + 1)[::-1]:

            # Determine the radix qubits to perform the addition on
            a_block = a[R * i : R * (i + 1)]
            b_block = b[R * i : R * (i + 1)]

            if check_if_fresh([c[i - 1]], c[i - 1].qs(), ignore_q_envs=False) or i == 0:
                c_in = None
                if a_block == "0" * len(a_block):
                    continue
            else:
                c_in = c[i - 1]

            if i == len(c):
                c_out = None
            else:
                c_out = c[i]

            if a_block != len(a_block) * "0":
                cq_gidney_adder(a_block, b_block, c_in=c_in, ctrl=ctrl, c_out=c_out)
            else:
                lin_incr(b_block, c_in=c_in, c_out=c_out)
